package example.figures;

import java.util.HashSet;
import java.util.Set;

import edu.neu.ccs.demeter.dj.ClassGraph;
import edu.neu.ccs.demeter.dj.Strategy;
import edu.neu.ccs.demeter.dj.TraversalGraph;
import edu.neu.ccs.demeter.dj.Visitor;

class ElementCounter extends Visitor {
	int count = 0;
	Set visited = new HashSet();					//There is a feature in DJ that automatically
	public void before(FigureElement fe) {		//gets elements via getter methods.  When a
		if (!visited.contains(fe)) {				//class has an instance variable and a getter
			visited.add(fe);						//method, referenced objects might be visited
			System.out.println("visiting " + fe);	//twice.  Therefore, it is the visitor's 
		}											//responsibility not to count duplicated
	}												//visits.
	public int getCount() {
		return visited.size();
	}
}

public class CountElements {

	public static void main(String[] args) {
		Display disp = new Display();
		Figure fig = new Figure();

		Point p1 = fig.makePoint(disp, 0, 0);
		Point p2 = fig.makePoint(disp, 50, 50);
		Line l1 = fig.makeLine(disp, 100, 50, 150, 100);
		Line l2 = fig.makeLine(disp, 150, 50, 100, 100);

		fig.dump();

		System.out.println("Counting figure elements:");

		ClassGraph cgraph = new ClassGraph("example.figures");
		// System.out.println("\nclass graph:\n" + cg);

		Strategy sg =
			new Strategy("from example.figures.Figure to example.figures.FigureElement");

		TraversalGraph tgraph = new TraversalGraph(sg, cgraph);
		// System.out.println("\ntraversal graph:\n" + tgraph);

		ElementCounter visitor = new ElementCounter();

		tgraph.traverse(fig, visitor);

		System.out.println("there are " + visitor.getCount() + " elements.");

		// NOTE: some lessons from DJ
		// * objects in List may not be visited properly, workaround is to
		//   define a method that returns an array of FigureElement.
		//
		// * instance variable named "elements" seems to be magical word for DJ,
		//   as it caused ClassCastException upon traversal.  Workaround is,
		//   unfortunately, to rename the variable.
		//
		// * each point in a line is visited twice (probably) because
		//   DJ follows the instance variables and getter methods.
	}
}