#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <2002-06-11 12:58:15 gregor>"))
  (display (string-append "ast.scm: " (substring time-stamp 13 29)))
  (newline))

;;;
;;; conventions:
;;;
;;;  Anything that can be represented as a symbol, number or list is
;;;  represented that way.  BUT, variables/fields/accessors that hold/access
;;;  such values are never named by the representation, they are named by
;;;  what is represented.  For example:
;;;
;;;      the 'name' of a field
;;;      the 'params' of a method
;;;      the 'exps' of a body
;;;
;;; See naming.txt for more on this.
;;;


;;;
;;; programs
;;;
(define-struct program (decls))

;;;
;;; decls
;;;
(define-struct declaration ())

(define-struct (class-decl     struct:declaration) (cname sname decls))       ;base
					                                     ;
(define-struct (member-decl    struct:declaration) (cname))                   ;
(define-struct (method-decl    struct:member-decl) (rtype mname params body)) ;
(define-struct (field-decl     struct:member-decl) (type fname))              ;

(define-struct (advice-decl    struct:member-decl) (qualifier pcd body))      ;ajd
(define-struct (named-pcd-decl struct:member-decl) (name params pcd))         ;

(define is-around? (lambda (adecl) (eq? (advice-decl-qualifier adecl) 'around)))
(define is-before? (lambda (adecl) (eq? (advice-decl-qualifier adecl) 'before)))
(define is-after?  (lambda (adecl) (eq? (advice-decl-qualifier adecl) 'after)))

;;;
;;; method (also advice, ...) bodies
;;;
(define-struct body (exps))

;;;
;;; params
;;;
(define-struct param (type pname))	;!!! type should be texp, ignore for now

(define params->types  (lambda (params) (map param-type  params)))
(define params->pnames (lambda (params) (map param-pname params)))

;;;
;;; expressions
;;;
(define-struct expression ())

(define-struct (lit-exp         struct:expression) (datum))
(define-struct (prim-app-exp    struct:expression) (prim rands))
(define-struct (if-exp          struct:expression) (test-exp then-exp else-exp))
(define-struct (and-exp         struct:expression) (rands))
(define-struct (or-exp          struct:expression) (rands))

(define-struct (let-exp         struct:expression) (params rands body))
(define-struct (var-get-exp     struct:expression) (vname))
(define-struct (var-set-exp     struct:expression) (vname rhs-exp))

(define-struct (field-get-exp   struct:expression) (obj-exp fname))
(define-struct (field-set-exp   struct:expression) (obj-exp fname rhs-exp))
(define-struct (method-call-exp struct:expression) (obj-exp mname rands))
(define-struct (super-call-exp  struct:expression) (mname rands))
(define-struct (new-obj-exp     struct:expression) (cname rands))
(define-struct (instanceof-exp  struct:expression) (cname obj-exp))

;;;
;;; aspects
;;;
(define-struct (proceed-exp struct:expression) (rands))

;;;
;;; pointcut designators
;;;
(define-struct pcd ())

(define-struct (call-pcd       struct:pcd) (rtype cname mname params))
(define-struct (mexecution-pcd struct:pcd) (rtype cname mname params))
(define-struct (aexecution-pcd struct:pcd) ())

(define-struct (new-pcd        struct:pcd) (cname params))

(define-struct (get-pcd        struct:pcd) (fname))
(define-struct (set-pcd        struct:pcd) (fname))

(define-struct (cwithin-pcd    struct:pcd) (cname))
(define-struct (mwithin-pcd    struct:pcd) (mname))

(define-struct (this-pcd       struct:pcd) (param))
(define-struct (target-pcd     struct:pcd) (param))
(define-struct (args-pcd       struct:pcd) (params))

(define-struct (named-pcd      struct:pcd) (name params))

(define-struct (cflow-pcd      struct:pcd) (pcd))
(define-struct (cflowtop-pcd   struct:pcd) (pcd))
(define-struct (cflowbelow-pcd struct:pcd) (pcd))

(define-struct (morerecent-pcd struct:pcd) (pcd-a pcd-b))

(define-struct (traversal-pcd  struct:pcd) (name))
(define-struct (crossing-pcd   struct:pcd) (fname))

(define-struct (and-pcd        struct:pcd) (pcds))
(define-struct (or-pcd         struct:pcd) (pcds))
(define-struct (not-pcd        struct:pcd) (pcd))
