#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <2002-06-11 12:59:15 gregor>"))
  (display (string-append "base.scm: " (substring time-stamp 13 29)))
  (newline))

(define run
  (lambda (text)
    (if (program? text)			;as a special favor to the hyperspaces code,
	(eval-program text)             ;we can take a program instead of text
	(eval-program (parse-program text)))))

(define eval-program
  (lambda (pgm)
    (set! *classes* (init-classes))
    (elaborate-class-decls! (program-decls pgm))
    (eval-exp
      (parse-exp `(send (new main) main))
      (make-empty-env))))

(define-struct class  (cname sname fields methods))
(define-struct method (cname mname params body))
(define-struct field  (cname type fname))

(define *classes* '())

(define init-classes
  (lambda () 
    (let ((class-object (make-class 'object #f '() (list (make-method 'object 'init '()
                                                           (parse-body '(1))))))
          (class-int    (make-class 'int    #f '() '()))
          (class-string (make-class 'string #f '() '())))
      (list class-object class-int class-string))))


(define elaborate-class-decls!
  (lambda (decls)
    (for-each (lambda (decl)
		(let* ((cname (class-decl-cname decl))
		       (sname (class-decl-sname decl))
		       (fdecls (collect-if  field-decl? (class-decl-decls decl)))
		       (mdecls (collect-if method-decl? (class-decl-decls decl)))
		       (fields  (roll-up-fields  cname sname fdecls))
		       (methods (roll-up-methods cname sname mdecls))
		       (class (make-class cname sname fields methods)))
		  (set! *classes* (append *classes* (list class)))
		  class))
	      decls)))

(define roll-up-fields
  (lambda (cname sname fdecls)
    (let ((super-fields (cname->fields sname))		
	  (  our-fields (map (lambda (fdecl)
			       (make-field cname
					   (field-decl-type fdecl)
					   (field-decl-fname fdecl)))
			     fdecls)))
      (append-keep-first-if (lambda (fa fb)
			      (eqv? (field-fname fa) (field-fname fb)))
			    super-fields our-fields))))

(define roll-up-methods
  (lambda (cname sname mdecls)
    (letrec ((super-methods (cname->methods sname))
             (  our-methods (map (lambda (mdecl)
                                   (make-method cname 
                                                (method-decl-mname mdecl)
                                                (method-decl-params mdecl)
                                                (method-decl-body mdecl)))
                                 mdecls)))
      (append-keep-first-if (lambda (ma mb)
                              (eqv? (method-mname ma) (method-mname mb)))
                            our-methods
                            super-methods))))


;;;;;;;;;;;;;;;; the interpreter ;;;;;;;;;;;;;;;;

(define eval-exp
  (lambda (exp env)
    (cond ((lit-exp? exp) (lit-exp-datum exp))
          
          ((if-exp? exp)
           (if (eval-exp (if-exp-test-exp exp) env)
               (eval-exp (if-exp-then-exp exp) env)
               (eval-exp (if-exp-else-exp exp) env)))
     
          ((prim-app-exp? exp)
           (let ((args (eval-rands (prim-app-exp-rands exp) env)))
             (apply-primitive (prim-app-exp-prim exp) args)))
     
          ((and-exp? exp)
           (let loop ((rands (and-exp-rands exp)))
             (if (null? rands)
                 #t
                 (and (eval-exp (car rands) env)
                      (loop (cdr rands))))))
     
          ((or-exp? exp)
           (let loop ((rands (or-exp-rands exp)))
             (if (null? rands)
                 #f
                 (or (eval-exp (car rands) env)
                     (loop (cdr rands))))))
     
          ((let-exp? exp)
           (let ((pnames (params->pnames (let-exp-params exp)))
                 (args   (eval-rands (let-exp-rands exp) env)))
             (eval-body (let-exp-body exp) (extend-env pnames args env))))
     
          ((var-get-exp? exp)
           (let ((vname  (var-get-exp-vname exp)))
             (deref (apply-env-ref env vname))))
     
          ((var-set-exp? exp)
           (let ((vname (var-set-exp-vname exp))
                 (nval  (eval-exp (var-set-exp-rhs-exp exp) env)))
             (setref! (apply-env-ref env vname) nval)
             'undefined))

          ((method-call-exp? exp)
           (call-method (method-call-exp-mname exp)
                        (eval-exp (method-call-exp-obj-exp exp) env)
                        (eval-rands (method-call-exp-rands exp) env)))
     
          ((super-call-exp? exp)
           (let* ((mname  (super-call-exp-mname exp))
                  (rands  (super-call-exp-rands exp))
                  (method (deref (apply-env-ref env '%method)))
                  (obj    (deref (apply-env-ref env 'this)))
                  (args   (eval-rands rands env)))
             (call-super-method mname (method->sname method) obj args)))
          
          ((new-obj-exp? exp)
           (new-object (new-obj-exp-cname exp)
                       (eval-rands (new-obj-exp-rands exp) env)))


          ((field-get-exp? exp)
           (let* (;(obj    (eval-exp (field-get-exp-obj-exp exp) env))
		  (obj    (deref (apply-env-ref env 'this)))
                  (method (deref (apply-env-ref env '%method)))
                  (fname  (field-get-exp-fname exp)))
             (get-field-value fname (method-cname method) obj)))

          ((field-set-exp? exp)
           (let* (;(obj    (eval-exp (field-set-exp-obj-exp exp) env))
		  (obj    (deref (apply-env-ref env 'this)))
                  (method (deref (apply-env-ref env '%method)))
                  (fname  (field-set-exp-fname exp))
                  (nval   (eval-exp (field-set-exp-rhs-exp exp) env)))
             (set-field-value! fname (method-cname method) obj nval)
             'undefined))

          ((instanceof-exp? exp)
           (instanceof? (eval-exp (instanceof-exp-obj-exp exp) env)
                         (instanceof-exp-cname exp)))
     
          (else (error 'eval-exp "not an expression -- ~s" exp)))))

(define eval-body
  (lambda (body env)
    (let loop ((exps (body-exps body)) (val #f))
      (if (null? exps)
          val
          (loop (cdr exps) (eval-exp (car exps) env))))))

(define eval-rands
  (lambda (rands env)
    (map-lr (lambda (x) (eval-exp x env)) rands)))




(define lookup-and-execute-method
  (lambda (mname obj args)
    (execute-method
      (lookup-method mname (object-cname obj)) obj args)))
(define call-method lookup-and-execute-method)

(define call-super-method
  (lambda (mname sname obj args)
    (execute-method
      (lookup-method mname sname) obj args)))

(define execute-method
  (lambda (method this args)
    (if (procedure? (method-body method))             ;a native method
        (apply (method-body method) (cons this args)) ;escape to Scheme
        (let ((pnames (method->pnames method)))
          (eval-body (method-body method)
            (new-env (cons 'this (cons '%method pnames))
                     (cons  this (cons   method args))))))))


(define new-object
  (lambda (cname args)
    (let* ((fnames (cname->fnames cname))
           (vals   (make-vector (length fnames)))
           (obj    (make-object cname vals)))
      (execute-method (lookup-method 'init cname) obj args) ;no mcall jp for
                                                            ;the call to init
      obj)))




(define get-field-value
  (lambda (fname within-method-cname obj)
    (deref (lookup-field-ref fname within-method-cname obj))))

(define set-field-value!
  (lambda (fname within-method-cname obj nval)
    (setref! (lookup-field-ref fname within-method-cname obj) nval)))



(define lookup-class 
  (lambda (cname)
    (or (find-if (lambda (class) (eqv? (class-cname class) cname)) *classes*)
        (error 'lookup-class "No class named ~s." cname))))

(define lookup-method
  (lambda (mname cname)
    (or (find-if (lambda (method) (eqv? (method-mname method) mname))
                 (cname->methods cname))
        (error 'eval-expr "No method for ~s in ~s." mname cname))))

(define lookup-field-ref 
  (lambda (fname cname object)
    (let* ((fnames (cname->fnames cname))
           (pos    (list-find-last-position fname fnames)))
      (if (number? pos) 
          (make-ref pos (object-vals object))
          (error 'lookup-field-pos
                 "~s does not have a field ~s." cname fname)))))



;;;;;;;;;;;;;;;; expressed values ;;;;;;;;;;;;;;;;

;;; see note in parser-scm

;;;;;;;;;;;;;;;; primitives ;;;;;;;;;;;;;;;;

(define apply-primitive
  (lambda (prim args)
    (apply prim args))) ; we just absorb our primitives from Scheme


;;;;;;;;;;;;;;;; objects ;;;;;;;;;;;;;;;;;;;;;

(define-struct object (cname vals))

(define cname-of
  (lambda (x)
    (cond ((object? x)  (object-cname x))
          ((integer? x) 'int)
          ((string? x)  'string)
          (else 
           (error 'cname-of "What is the class of ~s." x)))))


(define instanceof?
  (lambda (obj cname)
    (subclass? (object-cname obj) cname)))

(define subclass?
  (lambda (subcname supercname)
    (cond ((eqv? subcname supercname) #t)
          ((not subcname) #f)
          (else (subclass? (cname->sname subcname) supercname)))))


(define cname->fields
  (lambda (cn) (class-fields (lookup-class cn))))
(define cname->fnames
  (lambda (cn) (map field-fname (cname->fields cn))))

(define cname->sname
  (lambda (cname)
    (class-sname (lookup-class cname))))

(define cname->methods 
  (lambda (cname)
    (class-methods (lookup-class cname))))

(define method->pnames
  (lambda (method)
    (params->pnames (method-params method))))

(define method->sname
  (lambda (method)
    (cname->sname (method-cname method))))

;;;;;;;;;;;;;;;; environments ;;;;;;;;;;;;;;;;

(define-struct environment ())
(define-struct (empty-env    struct:environment) ())
(define-struct (extended-env struct:environment) (ids vals env))

(define-struct ref (pos vec))

(define extend-env
  (lambda (ids vals env)
    (make-extended-env ids (list->vector vals) env)))

(define new-env
  (lambda (ids vals)
    (extend-env ids vals (make-empty-env))))

(define apply-env-ref
  (lambda (env id)
    (cond ((empty-env? env) (error 'apply-env-ref "No binding for ~s" id))

          ((extended-env? env)
           (let ((pos (rib-find-position id (extended-env-ids env))))
             (if (number? pos)
                 (make-ref pos (extended-env-vals env))
                 (apply-env-ref (extended-env-env env) id))))
          
          (else (error 'apply-env "not an environment -- ~s" env)))))


(define deref
  (lambda (ref)
    (vector-ref (ref-vec ref) (ref-pos ref))))

(define setref!
  (lambda (ref val)
    (vector-set! (ref-vec ref) (ref-pos ref) val)
    #t))

(define rib-find-position
  (lambda (id los)
    (list-find-position id los)))
