#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <2001-11-09 14:52:16 gregor>"))
  (display (string-append "misc.scm:   " (substring time-stamp 13 29)))
  (newline))

;;;;;;;;;;;;;;;; misc.scm ;;;;;;;;;;;;;;;;

(define *require-db* '())

(define require
  (lambda (filename)
    (if (not (member filename *require-db*))
	(begin
	  (load filename)
	  (set! *require-db* (cons filename *require-db*))))))


(define caddddr (lambda (x) (car (cddddr x))))

(define map-lr
  (lambda (proc l)
    (if (null? l) 
	'()
	(cons (proc (car l)) 
	      (map-lr proc (cdr l))))))

(define last
  (lambda (lst)
    (cond ((null? lst) #f)
	  ((null? (cdr lst)) lst)
	  (else (last (cdr lst))))))


(define remove
  (lambda (item lst)
    (remove-if (lambda (x) (eqv? x item)) lst)))

(define remove*
  (lambda (item lst)
    (remove*-if (lambda (x) (eqv? x item)) lst)))

(define remove-if
  (lambda (pred lst)
    (if (null? lst)
	'()
        (if (pred (car lst))
            (cdr lst)
            (cons (car lst) (remove-if pred (cdr lst)))))))

(define remove*-if
  (lambda (pred lst)
    (if (null? lst)
	'()
        (if (pred (car lst))
            (remove*-if pred (cdr lst))
            (cons (car lst) (remove*-if pred (cdr lst)))))))

(define remove-duplicates
  (lambda (lst)
    (if (null? lst) '()
	(if (member (car lst) (cdr lst))
	    (remove-duplicates (cdr lst))
	    (cons (car lst) (remove-duplicates (cdr lst)))))))

(define remove-duplicates-if
  (lambda (binary-pred lst)
    (let ((pred (lambda (elm) (binary-pred (car lst) elm))))
      (let loop ((lst lst))
	(if (null? lst)
	    '()
	    (if (member-if pred (cdr lst))
		(remove-duplicates-if binary-pred (cdr lst))
		(cons (car lst) (remove-duplicates-if binary-pred (cdr lst)))))))))

(define append-keep-first-if
  (lambda (pred l1 l2)
    (letrec ((helper
	       (lambda (t2)
		 (if (null? t2) '()
		     (if (member-if (lambda (elt) (pred (car t2) elt)) l1)
			 (helper (cdr t2))
			 (cons (car t2) (helper (cdr t2))))))))
      (append l1 (helper l2)))))

(define member-if
  (lambda (pred lst)
    (if (null? lst) #f
	(if (pred (car lst)) lst
	    (member-if pred (cdr lst))))))


(define list-find-position
  (lambda (sym los)
    (let loop ((tail los) (pos 0))
      (cond ((null? tail) #f)
	    ((eqv? sym (car tail)) pos)
	    (else (loop (cdr tail) (+ pos 1)))))))

(define list-find-last-position
  (lambda (sym los)
    (let loop
	((los los) (curpos 0) (lastpos #f))
      (cond
       ((null? los) lastpos)
       ((eqv? sym (car los))
	(loop (cdr los) (+ curpos 1) curpos))
       (else (loop (cdr los) (+ curpos 1) lastpos))))))


(define every?
  (lambda (pred . lsts)
    (if (null? (car lsts)) #t
	(and (apply pred        (map car lsts))
	     (apply every? pred (map cdr lsts))))))

(define some?
  (lambda (pred . lsts)
    (if (null? (car lsts)) #f
	(or (apply pred       (map car lsts))
	    (apply some? pred (map cdr lsts))))))

; part of R5RS
; (define for-each
;   (lambda (fn lst)
;     (if (null? lst) #t
; 	(begin
; 	  (fn (car lst))
; 	  (for-each fn (cdr lst))))))

(define find-if
  (lambda (pred lst)
    (cond ((null? lst)      #f)
	  ((pred (car lst)) (car lst))
	  (else             (find-if pred (cdr lst))))))

(define collect-if
  (lambda (pred lst)
    (cond ((null? lst)      '())
	  ((pred (car lst)) (cons (car lst) (collect-if pred (cdr lst))))
	  (else             (collect-if pred (cdr lst))))))

(define collect-if-not
  (lambda (pred lst)
    (collect-if (lambda (x) (not (pred x))) lst)))

(define find-best
  (lambda (pred lst)
    (cond ((null? lst) #f)
	  ((null? (cdr lst)) (car lst))
	  (else
	   (let loop ((tail        (cdr lst))
		      (best-so-far (car lst)))
	     (cond ((null? tail) best-so-far)
		   ((pred (car tail) best-so-far)
		    (loop (cdr tail) (car tail)))
		   (else
		    (loop (cdr tail) best-so-far))))))))


(define make-list
  (lambda (length elt)
    (if (= length 0) 
	'()
	(cons elt (make-list (- length 1) elt)))))


(define iota
  (lambda (end)
    (let loop ((next 0))
      (if (>= next end) '()
	  (cons next (loop (+ 1 next)))))))

(define identity (lambda (x) x))

(define symbol-append
  (lambda parts
    (string->symbol (apply string-append (map symbol->string parts)))))



(define copies
  (lambda (n item)
    (let loop ((n n))
      (if (= 0 n)
	  '()
	  (cons item (loop (- n 1)))))))

(define merge
  (lambda lols
    (cond ((null? lols)       '())
	  ((null? (car lols)) '())
	  (else (append (map car lols)
			(apply merge (map cdr lols)))))))

(define firstn
  (lambda (lst n)
    (if (= n 0) '()
	(cons (car lst) (firstn (cdr lst) (- n 1))))))

(define subst-at
  (lambda (val index lst)
    (if (= index 0) (cons val (cdr lst))
	(cons (car lst) (subst-at val (- index 1) (cdr lst))))))


(define ignore
  (lambda args '())) ;oh, days gone by...


(define flatmap
  (lambda (f elements)
    (flatten (map f elements))))

(define flatten
  (lambda (list-of-lists)
    (apply append list-of-lists)))

(define filter-map
  (lambda (proc . lists)
    (remove* #f (apply map proc lists))))

;;; (fold f (list v1 v2 ... vN) unit)
;;; = (f vN .. (f v2 (f v1 unit))..)
(define fold
  (lambda (op elms unit)
    (if (null? elms)
	unit
	(fold op (cdr elms) (op (car elms) unit)))))

(define subset?
  (lambda (set1 set2)
    (every? (lambda (element1) (memq element1 set2))
	    set1)))

(define set-join
  (lambda set-list
    (remove-duplicates (apply append set-list))))

(define set-difference
  (lambda (set1 set2)
    (collect-if-not (lambda (elm) (memq elm set2))
		    set1)))

(define powerset			; all subsets of a set
  (lambda (set)
    (if (null? set)
	'(())
	(let ((element           (car set))
	      (powerset-of-rests (powerset (cdr set))))
	  (append (map (lambda (subset) (cons element subset))
		       powerset-of-rests)
		  powerset-of-rests)))))

;;;
;;; (set! foo
;;;   (around-wrap foo
;;;     (lambda (original-foo . args)
;;;       <code here wraps foo and call call it with original-foo>
;;;       )))
;;;

(define around-wrap
  (lambda (original-proc around)
    (lambda args
      (apply around (cons original-proc args)))))

(define before-wrap
  (lambda (proc before)
    (around-wrap proc
      (lambda (original-proc . args)
	(apply before  args)
	(apply original-proc args)))))

(define after-wrap
  (lambda (proc after)
    (around-wrap proc
      (lambda (original-proc . args)
	(let ((result (apply original-proc args)))
	  (apply after  args)
	  result)))))
