#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <2002-02-18 13:12:53 gregor>"))
  (display (string-append "tests=.scm:  " (substring time-stamp 13 29)))
  (newline))

(define *tests* '()) ; ((id desc pgm value msgs) ...)

(define add-test
  (lambda (id desc value pgm msgs)
    (let ((existing (assq id *tests*))
	  (value    (list  desc value pgm msgs)))
      (if existing
	  (set-cdr! existing value)
	  (set! *tests* (append *tests* (list (cons id value))))))))

(define run-tests
  (lambda ()
    (for-each (lambda (test) (apply run-test-1 test)) *tests*)))

(define run-test
  (lambda (id . debug)
    (let ((test (assq id *tests*)))
      (cond ((eqv? test #f)
	     (error 'run-tests "No test named ~s." id))
	    ((null? debug)
	     (apply run-test-1 test))
	    (else
	     (let ((pgm (list-ref test 2)))
	       (print (run pgm))
	       (newline)))))))

(define run-test-1
  (lambda (id desc pgm value msgs)
    (let ((real-error   error)
	  (caught-error #f)
	  (return-msgs  '()))
      (let ((return-value
	     (call-with-current-continuation
	      (lambda (abort)
		(fluid-let ((error	;NOTE: format and error are not R5RS
			     (lambda args
			       (set! error real-error)
			       (set! caught-error args)
			       (abort 'error)))
			    (write
			     (lambda x
			       (set! return-msgs
				     (append return-msgs
					     x))))
			    (newline
			     (lambda ()
			       (set! return-msgs
				     (append return-msgs
					     (list 'NL)))))
			    )
		  (run pgm))))))
	(let ((width 16)
	      (slen (string-length (symbol->string id))))
	  (if (< slen width)
	      (display (make-string (- width slen) #\space))))
	(display id)
	(display ": ")
	(cond ((and (equal? value 'error)
		    (equal? msgs caught-error))
		(display "Succeeded (failed as expected). "))
	      ((and (equal? value 'error) 
		    caught-error)
	       (display msgs)
	       (display caught-error)
	       (display "FAILED    - by failing with wrong error message. ")
	       (display caught-error))
	      ((and (equal? value 'error) (not caught-error))
	       (display "FAILED    - by not failing. "))
	      (caught-error
	       (display "FAILED    - by failing unexpectedly. ")
	       (display caught-error))
	      ((and (equal? value return-value)
		    (equal? msgs return-msgs))
	       (display "Succeeded (correct result and messages). "))
	      (else 
	       (display "FAILED    - ")
	       (if (not (equal? value return-value))
		   (display "incorrect result "))
	       (if (not (equal? msgs  return-msgs))
		   (display "incorrect messages "))))
	(display "[")
	(display desc)
	(display "] ")
	(newline)))))

;;;;;;;;;;;;;;;; tests ;;;;;;;;;;;;;;;;

(define main-w-fact-plus-member-decls
  (lambda (decls)
    `((class main object
	(method void init () (super init))
	(method int main () (send this fact 6))
	(method int fact ((int n))
	  (if (< n 1) 1
	      (* n (send this fact (- n 1)))))
	,@decls))))

(define main-w-employee-plus-class-decls
  (lambda (decls)
    `((class main object
	(method void init () (super init))
	(method int main ()
	  (let (((employee john) (new employee "john"))
		((employee bob)  (new employee "bob")))
	    (send john pay)
	    (send bob pay))))

      (class employee object
	(field string name)
	(method void init ((string name))
	  (super init)
	  (fset! this name name))

	(method void print () (write (fget this name)))

	(field int salary)
	(method int pay ()
	  (write 'paying:) (send this print) (newline)
	  (fget this salary)))
      
      ,@decls)))

(add-test 'base01
	  "starting main method"
	  '((class main object
	      (method void init ())
	      (method void main () (+ 1 3))))
	  4
	  '())

(add-test 'base02
	  "error messages match"
	  '((class main object
	      (method void init ())
	      (method int main () (+ 1 a))))
	  'error
	  '(apply-env-ref "No binding for ~s" a))

(add-test 'base03
	  "method call"
	  (main-w-fact-plus-member-decls '())
	  720
	  '())

(add-test 'base04
	  "more oo"
	  (main-w-employee-plus-class-decls '())
	  0
	  '(paying: "john" NL paying: "bob" NL))


#|  !!! make these tests work !!!
(add-test 'base05
	  "dynamic dispatch"
	  '((class main object
	      (method void init ())
	      (method void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2)))
				      (+ (* 2 (send o1 foo))
					 (send o2 foo)))))
	    (class c1 object
		   (method int foo   () 5)
		   (method void init () 1))
	    (class c2 object
		   (method int foo   () 10)
		   (method void init () 2)))
	  20
	  '())

(add-test 'base06
	  "super dispatch and inheritance"
	  '((class main object
	      (method void init ())
	      (method void main () (let (((c1 o1) (new c1 1))
					 ((c2 o2) (new c2 2)))
				      (+ (send o1 foo o1)
					 (send o1 foo o2)
					 (send o2 foo o2)))))

	    ; remember: no field shadowing
	    (class c1 object
		   (field int y)
		   (method int foo   ((c1 c))   (send c bar))
		   (method int bar   ()         (fget this y))
		   (method void init ((int ty)) (fset! this y ty)))

	    (class c2 c1
		   (field int x)

		   (method int bar   ()
		     (+ (fget this x) (fget this y)))

		   (method void init ((int tx))
		     (super init (* 2 tx)) (fset! this x tx))))
	  13
	  '())

|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; example : figures

(define figures-with-decls
  (lambda (class-decls . main-statements)
    `((class main object
	     (method void main ()
	       ,@(if (null? main-statements)
		     `((let (((figure fig)  (new figure))
			     ((display disp) (new display)))

			 (let (((point  p1)  (send fig make-point disp  0  0))
			       ((point  p2)  (send fig make-point disp 50 50))
			       ((line   l1)  (send fig make-line  disp 100 50 150 100))
			       ((line   l2)  (send fig make-line  disp 150 50 100 100)))

			   (send fig dump)

			   (write "moving elements")(newline)
			   (send p1 set-x 30)
			   (send l2 set-p2 p2)

			   (send fig dump))))
		     main-statements)))

      (class figure object
	     (field linked-list element-list)
	     (method void init () (super init)
		     (fset! this element-list '()))

	     (method void dump ()
		     (write "figure:") (newline)
		     (send this dump-elements (fget this element-list)))
	     (method void dump-elements ((list elements))
		     (if (null? elements)
			 'done
			 (let ()
			   (send (send elements car)
				 print)
			   (newline)
			   (send this dump-elements
				 (send elements cdr)))))

	     (method point make-point ((display display) (int x) (int y))
		     (let (((point p) (new point display x y)))
		       (fset! this element-list
			    (new linked-list p (fget this element-list))
			    )		       
		       p))

	     (method void make-line ((display display) (int x1) (int y1) (int x2) (int y2))
		     (let (((point p1) (new point display x1 y1))
			   ((point p2) (new point display x2 y2)))
		       (let (((line l) (new line display p1 p2)))
			 (fset! this element-list
				(new linked-list l (fget this element-list))
				)		       
			 l)))
	     )

      (class linked-list object
	     (field figure-element car)
	     (field linked-list cdr)
	     (method void init ((figure-element car) (linked-list cdr))
		     (super init)
		     (fset! this car car)
		     (fset! this cdr cdr))
	     (method figure-element car () (fget this car))
	     (method figure-element cdr () (fget this cdr)))

      (class display object
 	     (method void update ((figure-element fe))
		     (write 'update:) (send fe print) (newline))

	     (method void draw-point ((int x) (int y))
		     (write (list 'draw-point x y)) (newline))
	     (method void draw-line ((int x1) (int y1) (int x2) (int y2))
		     (write (list 'draw-line x1 y1 x2 y2)) (newline))
	     (method void draw-oval ((int x1) (int y1) (int x2) (int y2))
		     (write (list 'draw-oval x1 y1 x2 y2)) (newline))
	     )

      (class figure-element object
	     (field display display)
	     (method void init ((display display)) 
	       (super init)
	       (fset! this display display))
	     (method display get-display ()
		     (fget this display))
	     (method void move-by ((int x) (int y)) 'done)
	     )
      
      (class point figure-element
	     (method void init ((display display) (int x) (int y)) 
		     (super init display)
		     (fset! this x x)
		     (fset! this y y))
	     (field int x)
	     (field int y)
	     (method int get-x () (fget this x))
	     (method int get-y () (fget this y))
	     (method void set-x ((int x)) (fset! this x x))
	     (method void set-y ((int y)) (fset! this y y))
	     (method void move-by ((int x) (int y))
	       (send this set-x (+ x (send this get-x)))
	       (send this set-y (+ y (send this get-y))))

	     (method void print ()
		     (write 'Point<)
		     (write (send this get-x))
		     (write ':)
		     (write (send this get-y))
		     (write '>)
		     ))

      (class line figure-element
	     (method void init ((display display) (point p1) (point p2))
		     (super init display)
		     (fset! this p1 p1)
		     (fset! this p2 p2))
	     (field point p1)
	     (field point p2)
	     (method int get-p1 () (fget this p1))
	     (method int get-p2 () (fget this p2))
	     (method void set-p1 ((int p1)) (fset! this p1 p1))
	     (method void set-p2 ((int p2)) (fset! this p2 p2))
	     (method void move-by ((int x) (int y))
	       (send p1 move-by x y)
	       (send p2 move-by x y))

	     (method void print ()
		     (write 'line<)
		     (send (send this get-p1) print)
		     (write '-)
		     (send (send this get-p2) print)
		     (write '>)))
      ,@ class-decls
      )))

(define figures-with-member-decls
  (lambda (member-decls)
    (figures-with-decls
     `((class aspect object ,@member-decls)))))

(add-test 'base05
	  "figures"
	  (figures-with-decls '())
	  'done
	  '("figure:" NL
	    line< point< 150 : 50 > - point< 100 : 100 > > NL
	    line< point< 100 : 50 > - point< 150 : 100 > > NL
	    point< 50 : 50 > NL
	    point< 0 : 0 > NL
	    "moving elements" NL
	    "figure:" NL
	    line< point< 150 : 50 > - point<  50 : 50 > > NL
	    line< point< 100 : 50 > - point< 150 : 100 > > NL
	    point< 50 : 50 > NL
	    point< 30 : 0 > NL))
