#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <2002-02-18 13:13:09 gregor>"))
  (display (string-append "tests=compositor.scm:  " (substring time-stamp 13 29)))
  (newline))


(define main-program
  `((class main object
      (method void init () (super init))
      (method int main ()
        (let (((employee john) (new employee "john"))
              ((employee bob) (new employee "bob")))
          (send john add-manager bob)
          (send john add-manager john)
          
          (send john check) (send john pay)
          (send bob check) (send bob pay))))

    (class employee object

      (field string name)
      (field int salary)
      (field (list employee) managers)

      (method void init ((int name))
        (super init)
        (fset! this name name)
        (fset! this managers '()))

      (method void print () (write (fget this name)))

      (method void add-manager ((employee mgr))
        (fset! this managers (cons mgr (fget this managers))))
      (method void check ()
        (write 'checking:)
        (send this print)
        (let (((int n) (length (fget this managers))))
          (if (< n 1)
              (write ':too-few-managers)
              #f)
          (if (< 3 n)
              (write ':too-many-managers)
              #f))
        (newline))

      (method int pay ()
        (write 'paying:) (send this print) (newline)
	(fget this salary)))))

(define no-pay
  '((class employee object
      (method int pay () 0))))

(define one-manager
  '((class employee object
      (method void check ()
        (let (((int n) (length (fget this managers))))
          (write 'checking:)
          (send this print)
          (if (not (= n 1))
              (write ':must-have-one-manager)
              #f)
          (newline))))))

(define main-slices
  (HJ-slice:weave
   (parse-program main-program)
   (parse-concern-map '(((field salary) payroll-concern)
			((method pay)   payroll-concern)
			(default        personnel-concern)))))


(define payroll-slice   (find-slice 'payroll-concern   main-slices))
(define personnel-slice (find-slice 'personnel-concern main-slices))
(define nopay-slice
  (let ((slices
	 (HJ-slice:weave
	  (parse-program no-pay)
	  (parse-concern-map '((default no-pay))))))
    (find-slice 'no-pay slices)))

;; (define business-rule-slices
;;   (make-slices
;;     personnel-slice
;;     (parse-concern-map '(((method check) three-managers)
;; 			 (default        no-business-rule)))))

;;(define personnel-no-biz-rule-slice (find-slice 'no-business-rule business-rule-slices))
(define one-manager-slice
  (let ((slices
	 (HJ-slice:weave
	  (parse-program one-manager)
	  (parse-concern-map '((default one-manager))))))
    (find-slice 'one-manager slices)))


(add-test 'hs0
	  "entire payroll program"
	  main-program
	  0
	  '(checking: "john" NL
	    paying: "john" NL
	    checking: "bob" :too-few-managers NL
	    paying: "bob" NL
	    ))

(add-test 'hs1
	  "personnel plus payroll"
	  (compositor:weave (list personnel-slice payroll-slice)
				merge-by-name)
	  0
	  '(checking: "john" NL
	    paying: "john" NL
	    checking: "bob" :too-few-managers NL
	    paying: "bob" NL
	    ))

(add-test 'hs2
	  "personnel minus payroll"
	  (compositor:weave (list personnel-slice nopay-slice)
				merge-by-name)
	  0
	  '(checking: "john" NL
	    checking: "bob" :too-few-managers NL
	    ))

;; (add-test 'hs3
;; 	  "personnel plus hr"
;; 	  (make-program (merge-by-name (list personnel-no-biz-rule-slice 
;; 					     one-manager-slice 
;; 					     nopay-slice)))
;; 	  0
;; 	  '(checking: "john" :must-have-one-manager NL
;; 	    checking: "bob" :must-have-one-manager NL
;; 	    ))


(add-test 'hs4
	  "personnel plus hr"
	  (compositor:weave (list personnel-slice
				      one-manager-slice 
				      nopay-slice)
				merge-by-name)
	  0
	  '(checking: "john" NL
	    checking: "john" :must-have-one-manager NL
	    checking: "bob" :too-few-managers NL
	    checking: "bob" :must-have-one-manager NL
	    ))

(define simple-program
  '((class main object
	   (method void init () (super init))
	   (method void main ()
		   (write 'main) (newline)
		   (send (new foo) test)
		   0))
    (class foo object
	   (method void init () (super init))
	   (method void test ()
		   (write 'foo) (newline)))
    (class bar object
	   (method void init () (super init))
	   (method void baz ()
		   (write 'bar) (newline)
		   1))))


(add-test 'hs5
	  "no renaming"
	  (compositor:weave
	   (HJ-slice:weave (parse-program simple-program)
			   (parse-concern-map '((default        default))))
	   merge-by-name)
	  0
	  '(main NL foo NL))


(add-test 'hs6
	  "method renaming"
	  (compositor:weave
	   (HJ-slice:weave (parse-program simple-program)
			   (parse-concern-map '((default        default))))
	   (make-relationships (list (match-rename? '(method (bar baz) (foo test)))
				     match-byname?)
			       'merge))
	  0
	  '(main NL foo NL bar NL))

(add-test 'hs7
	  "method renaming"
	  (compositor:weave
	   (HJ-slice:weave (parse-program simple-program)
			   (parse-concern-map '((default        default))))
	   (make-relationships (list (match-rename? '(method (bar baz) (foo test)))
				     (match-rename? '(method (bar baz) (main main)))
				     match-byname?)
			       'merge))
	  1
	  '(main NL foo NL bar NL bar NL))

(define observable-program
  '((class observable object
	   (field display display)
	   (method void moved ()
		   (send (fget this display) update this)))))

; (define figure
;   (compositor:weave
;    (HJ-slice:weave (parse-program (figures-with-decls observable-program))
; 		   (parse-concern-map '((default        default))))
;    merge-by-name))

; (define monitoring-figure
;   (compositor:weave
;    (HJ-slice:weave (parse-program (figures-with-decls '()))
; 		   (parse-concern-map '((default        default))))
;    (make-relationships
;     (list (match-rename? '(method (observable moved) (point setx)))
; 	  (match-rename? '(method (observable moved) (point sety)))
; 	  (match-rename? '(method (observable moved) (line setp1)))
; 	  (match-rename? '(method (observable moved) (line setp2)))
; 	  match-byname?)
; 		       'merge)))
