#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <2002-02-18 13:13:09 gregor>"))
  (display (string-append "tests=paex.scm:  " (substring time-stamp 13 29)))
  (newline))

(load "tests=pa.scm")

;;;;;;;;;;;;;;;; tests ;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; get join points

(add-test 'get01
	  "field get"
	  '((class main object
	      (field int x)
	      (method void main ()
		(fset! this x 1)
		(fget this x))
	      (method void init () (super init))
	      (before (get x) (write 'get-x) (newline))
	      (before (get y) (write 'get-x) (newline))))
	  1
	  '(get-x NL))

(add-test 'get02
	  "target of get"
	  '((class main object
	      (field int x)
	      (method void main () (fset! this x 1) (fget this x))
	      (method void print () (write 'main) (newline))
	      (method void init () (super init))
	      (before (and (get x) (target (main o)))
		      (write 'get-x) (newline)
		      (send o print))))
	  1
	  '(get-x NL main NL))

(add-test 'target02
	  "polymorphic target"
	  '((class main object
	      (field int x)
	      (method int main ()
		(fset! this x 1)
		(+ (fget this x)
		   (send (new point 1 2) get-y)))
	      (method void print () (write 'main) (newline))
	      (method void init () (super init))
	      (before (and (target (object o))
			   (or (get x) (get y) (call get-x) (call get-y)))
		      (write 'target:)
		      (send o print)))
	    (class point main
	      (field int x)
	      (field int y)
	      (method void init ((int x) (int y))
		(super init)
		(fset! this x x)
		(fset! this y y))
	      (method void print () (write 'point) (newline))
	      (method int get-x () (fget this x))
	      (method int get-y () (fget this y))))
	  3
	  '(target: main NL		; target of (fget this x)
	    target: point NL		; target of (send ... get-x)
	    target: point NL))		; target of (fget this x)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; after advice

(add-test 'after01
	  "after advice"
	  (main-w-fact-plus-member-decls
	    '((after (call fact)   (write 'after) (newline))))
	  720
	  (merge (copies 7 'after)
		 (copies 7 'NL)))

(add-test 'after02
	  "multiple after advice"
	  (main-w-fact-plus-member-decls
	    '((after (call fact)  (write 'after1) (newline))
	      (after (call fact)  (write 'after2) (newline))))
	  720
	  (merge (copies 7 'after1)
		 (copies 7 'NL)
		 (copies 7 'after2)
		 (copies 7 'NL)))

(add-test 'after03
	  "before and after advice"
	  (main-w-fact-plus-member-decls
	    '((before (call fact) (write 'before) (newline))
	      (after  (call fact) (write 'after) (newline))))
	  720
	  (append (merge (copies 7 'before)
			 (copies 7 'NL))
		  (merge (copies 7 'after)
			 (copies 7 'NL))))

(add-test 'after04
	  "after and args"
	  (main-w-fact-plus-member-decls
	    '((after (and (call fact)
			  (args (int x)))
		(write 'after) (write x) (newline))))
	  720
	  (merge (copies 7 'after)
		 (iota 7)
		 (copies 7 'NL)))


(add-test 'after05
	  "before and after with args"
	  (main-w-fact-plus-member-decls
	    '((before (and (call fact)
			   (args (int x)))
		(write 'before) (write x) (newline))
	      (after  (and (call fact)
			   (args (int y)))
		(write 'after) (write y) (newline))))
	  720
	  (append (merge (copies 7 'before)
			 (reverse (iota 7))
			 (copies 7 'NL))
		  (merge (copies 7 'after)
			 (iota 7)
			 (copies 7 'NL))))
	  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; around

(add-test 'around01
	  "1 around,  w/o proceed"
	  (main-w-fact-plus-member-decls
	    '((around (call fact) 1)))
	    1
	  '())

(add-test 'around02
	  "1 around,  w/  proceed, no args"
	  (main-w-fact-plus-member-decls
	    '((around (call fact)
		(write 'around-start) (newline)
		(proceed))))
	  720
	  (merge (copies 7 'around-start)
		 (copies 7 'NL)))

(add-test 'around03
	  "2 arounds,  w/  proceed, no args"
	  (main-w-fact-plus-member-decls
	    '((around (call fact)
		(write 'around1-start) (newline)
		(let (( (int r) (proceed) ))
		  (write 'around1-end) (newline)
		r))
	      (around (call fact)
		(write 'around2-start) (newline)
		(let (( (int r) (proceed) ))
		  (write 'around2-end) (newline)
		  r))))
	  720
	  ;; a1-s NL a2-s NL (rep. 7 times) a2-e NL a1-e NL ...(rep. 7 times)
	  (append
	   (merge (merge (copies 7 'around1-start)
			 (copies 7 'around2-start))
		  (copies 14 'NL))
	   (merge (merge (copies 7 'around2-end)
			 (copies 7 'around1-end))
		  (copies 14 'NL))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; cflow

(add-test 'cflow01
	  "simple cflow"
	  (main-w-fact-plus-member-decls
	    '((before (and (cflow (call fact))
			   (call fact)
			   (args (int x)))
		(write x) (newline))))
	  720
	  (merge (reverse (iota 7))
		 (copies 7 'NL)))


(add-test 'cflow02
	  "simple cflow w/ values"
 	  (main-w-fact-plus-member-decls
	    '((before (and (cflow (and (call fact) (args (int topval))))
			   (call fact)
			   (args (int thisval)))
		(write topval) (write thisval))))
	  720
	  (merge (reverse (iota 7))
		 (reverse (iota 7))))

(add-test 'cflow03
	  "cflow(cflow( of fact ))"
	  (main-w-fact-plus-member-decls
	    '((before (and (cflow (cflow (call fact)))
			   (call fact)
			   (args (int x)))
		(write x) (newline))))
	  720
	  (merge (reverse (iota 7))
		 (copies 7 'NL)))




; (add-test 'advice02b
; 	  "multiple before advice"
; 	  (main-w-fact-plus-member-decls
; 	    '((pointcut at-fact () (call fact))
; 	      (before (at-fact)  (write 'before1) (newline))
; 	      (before (at-fact)  (write 'before2) (newline))))
; 	  720
; 	  (merge (copies 7 'before1)
; 		 (copies 7 'NL)
; 		 (copies 7 'before2)
; 		 (copies 7 'NL)))

; (add-test 'advice04
; 	  "multiple after advice"
; 	  (main-w-fact-plus-member-decls
; 	    '((after (call fact)  (write 'after1) (newline))
; 	      (after (call fact)  (write 'after2) (newline))))
; 	  720
; 	  (merge (copies 7 'after1)
; 		 (copies 7 'NL)
; 		 (copies 7 'after2)
; 		 (copies 7 'NL)))

; (add-test 'advice05
; 	  "before and after advice"
; 	  (main-w-fact-plus-member-decls
; 	    '((before (call fact) (write 'before) (newline))
; 	      (after  (call fact) (write 'after) (newline))))
; 	  720
; 	  (append (merge (copies 7 'before)
; 			 (copies 7 'NL))
; 		  (merge (copies 7 'after)
; 			 (copies 7 'NL))))



; (add-test 'advice06b
; 	  "before and args"
; 	  (main-w-fact-plus-member-decls
; 	    '((pointcut at-fact ((int a)) (and (call fact) (args (int a))))
; 	      (before (at-fact (int x))
; 		(write 'before) (write x) (newline))))
; 	  720
; 	  (merge (copies 7 'before)
; 		 (reverse (iota 7))
; 		 (copies 7 'NL)))

; (add-test 'advice07
; 	  "after and args"
; 	  (main-w-fact-plus-member-decls
; 	    '((after (and (call fact)
; 			  (args (int x)))
; 		(write 'after) (write x) (newline))))
; 	  720
; 	  (merge (copies 7 'after)
; 		 (iota 7)
; 		 (copies 7 'NL)))

; (add-test 'advice08
; 	  "before and after with args"
; 	  (main-w-fact-plus-member-decls
; 	    '((before (and (call fact)
; 			   (args (int x)))
; 		(write 'before) (write x) (newline))
; 	      (after  (and (call fact)
; 			   (args (int y)))
; 		(write 'after) (write y) (newline))))
; 	  720
; 	  (append (merge (copies 7 'before)
; 			 (reverse (iota 7))
; 			 (copies 7 'NL))
; 		  (merge (copies 7 'after)
; 			 (iota 7)
; 			 (copies 7 'NL))))

; (add-test 'args01
; 	  "just args"
;    	  (main-w-fact-plus-member-decls
; 	    '((before (and (args (int x)) (not (aexecution)))
; 		(write x) (newline))))
; 	  720
; 	  (merge (reverse (iota 7))
; 		 (copies 7 'NL)
; 		 (reverse (iota 7))
; 		 (copies 7 'NL)))



; (add-test 'around03
; 	  "1 around,  w/  proceed, pass object, change args"
; 	  (main-w-fact-plus-member-decls
; 	    '((around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(proceed m 0))))
; 	  1
; 	  '())


; (add-test 'around04
; 	  "1 around,  w/  proceed, and some output"
; 	  (main-w-fact-plus-member-decls
; 	    '((around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(write 'around-before) (newline)
;                 (write n)              (newline)
; 		(let (((int r) (proceed m n)))
; 		  (write 'around-after)  (newline)
; 		  r))))
; 	  720
; 	  (append (merge (copies 7 'around-before)
; 			 (copies 7 'NL)
; 			 (reverse (iota 7))
; 			 (copies 7 'NL))
; 		  (merge (copies 7 'around-after)
; 			 (copies 7 'NL))))

; (add-test 'around05
; 	  "2 arounds, w/  proceed, pass object, pass args"
; 	  (main-w-fact-plus-member-decls
; 	    '((around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(let (((int r) 0))
; 		  (write 'around1-before) (newline)
; 		  (set! r (proceed m n))
; 		  (write 'around1-after) (newline)
; 		  r))
; 	      (around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 	        (let (((int r) 0))
; 		  (write 'around2-before) (newline)
; 		  (set! r (proceed m n))
; 		  (write 'around2-after) (newline)
; 		  r))))
; 	  720
; 	  (append (merge (copies 7 'around1-before)
; 			 (copies 7 'NL)
; 			 (copies 7 'around2-before)
; 			 (copies 7 'NL))
; 		  (merge (copies 7 'around2-after)
; 			 (copies 7 'NL)
; 			 (copies 7 'around1-after)
; 			 (copies 7 'NL))))

; (add-test 'around06
; 	  "2 arounds, w/  proceed, pass object, change args"
; 	  (main-w-fact-plus-member-decls
; 	    '((around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(proceed m (- n 2))
; 	       ;(proceed m n)
; 		)
; 	      (around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(write n)
; 		(proceed m (- n 4))
; 	       ;(proceed m n)
; 		)))
; 	  1
; 	  '(4))

; (add-test 'around07
; 	  "around and before advice"
; 	  (main-w-fact-plus-member-decls
; 	    '((around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(let (((int r) 0))
; 		  (write 'around-before) (newline)
; 		  (set! r (proceed m n))
; 		  (write 'around-after) (newline)
; 		  r))
; 	      (before (and (call fact)
; 			   (args (int x)))
; 		(write 'before) (newline))
; 	      ))
; 	  720
; 	  (append (merge (copies 7 'around-before)
; 			 (copies 7 'NL)
; 			 (copies 7 'before)
; 			 (copies 7 'NL))
; 		  (merge (copies 7 'around-after)
; 			 (copies 7 'NL))))

; (add-test 'around08
; 	  "around and after advice"
; 	  (main-w-fact-plus-member-decls
; 	    '((around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(let (((int r) 0))
; 		  (write 'around-before) (newline)
; 		  (set! r (proceed m n))
; 		  (write 'around-after) (newline)
; 		  r))
; 	      (after (and (call fact)
; 			  (args (int x)))
; 		(write 'after) (newline))
; 	      ))
; 	  720
; 	  (append (merge (copies 7 'around-before)
; 			 (copies 7 'NL))
; 		  (merge (copies 7 'after)
; 			 (copies 7 'NL)
; 			 (copies 7 'around-after)
; 			 (copies 7 'NL))))


; (add-test 'around09
; 	  "around, before and after advice"
; 	  (main-w-fact-plus-member-decls
; 	    '((around (and (call fact)
; 			   (target (main m))
; 			   (args (int n)))
; 		(let (((int r) 0))
; 		  (write 'around-before) (newline)
; 		  (set! r (proceed m n))
; 		  (write 'around-after) (newline)
; 		  r))
; 	      (before (and (call fact)
; 			   (args (int x)))
; 		(write 'before) (newline))
; 	      (after (and (call fact)
; 			  (args (int x)))
; 		(write 'after) (newline))
; 	      ))
; 	  720
; 	  (append (merge (copies 7 'around-before)
; 			 (copies 7 'NL)
; 			 (copies 7 'before)
; 			 (copies 7 'NL))
; 		  (merge (copies 7 'after)
; 			 (copies 7 'NL)
; 			 (copies 7 'around-after)
; 			 (copies 7 'NL))))




; (add-test 'cflow02
; 	  "simple cflowbelow"
; 	  (main-w-fact-plus-member-decls
; 	    '((before (and (cflowbelow (call fact))
; 			   (call fact)
; 			   (args (int x)))
; 		(write x) (newline))))
; 	  720
; 	  (merge (reverse (iota 6))
; 		 (copies 6 'NL)))

; (add-test 'cflow03
; 	  "simple cflowtop"
; 	  (main-w-fact-plus-member-decls
; 	    '((before (and (cflowtop (call fact))
; 			   (call fact)
; 			   (args (int x)))
; 		(write x) (newline))))
; 	  720
; 	  '(6 NL))


; (add-test 'cflow05
; 	  "simple cflowbelow w/ values"
; 	  (main-w-fact-plus-member-decls
; 	    '((before (and (cflowbelow (and (call fact)
; 					    (args (int topval))))
; 			   (call fact)
; 			   (args (int thisval)))
; 		(write topval) (write thisval))))
; 	  720
; 	  (merge (firstn (reverse (iota 7)) 6)
; 		 (cdr (reverse (iota 7)))))


; (add-test 'cflow07
; 	  "cflow(cflowbelow( of fact ))"
; 	  (main-w-fact-plus-member-decls
; 	    '((before (and (cflow (cflowbelow (call fact)))
; 			   (call fact)
; 			   (args (int x)))
; 		(write x) (newline))))
; 	  720
; 	  (merge (reverse (iota 6))
; 		 (copies 6 'NL)))

; (add-test 'cflow08
; 	  "cflow(cflowbelow ( of fact )) w/ values"
; 	  (main-w-fact-plus-member-decls
; 	    '((before (and (cflow (cflowbelow (and (call fact)
; 						   (args (int topval)))))
; 			   (call fact)
; 			   (args (int thisval)))
; 		(write topval) (write thisval))))
; 	  720
; 	  (merge (firstn (reverse (iota 7)) 6)
; 		 (cdr (reverse (iota 7)))))


; (add-test 'cwithin01
; 	  "cwithin(main)"
; 	  (main-w-employee-plus-class-decls
; 	    '((class cwithin01 object 
; 		(before (cwithin main)
; 			(write 'foo)))))
; 	  0
; 	  '(foo foo foo foo foo paying: "john" NL
; 	    foo paying: "bob" NL
; 	    ))

; (add-test 'cwithin02
; 	  "!cwithin(cwithin02)"
; 	  (main-w-employee-plus-class-decls
; 	    '((class cwithin02 object 
; 		(before (not (cwithin cwithin02))
; 			(write 'x)))))
; 	  0
; 	  '(x x x x x x x x x x x x x x x paying: x x x "john" NL
; 	    x x x paying: x x x "bob" NL
; 	    x ))

; (add-test 'mwithin01
; 	  "mwithin"
; 	  `((class main object
; 	      (method void init () (super init))
; 	      (method int main ()
; 		(let (((c1 o1) (new c1))
; 		      ((c1 o2) (new c2)))
; 		  (+ (send o1 foo)
; 		     (send o2 foo)
; 		     (send o1 bar)
; 		     (send o2 bar)
; 		     (send o1 baz)
; 		     (send o2 baz)))))
	    
; 	    (class c1 object
; 	      (method void init () 1)
; 	      (method int foo   () 1)
; 	      (method int bar   () 1)
; 	      (method int baz   () 1))

; 	    (class c2 c1
; 	      (method void init () 2)
; 	      (method int foo   () (super foo))
; 	      (method int bar   () 2))

; 	    (class testing object

; 	      (before (and (cwithin c1) (mwithin foo))
; 		      (write 'c1-foo))
; 	      (before (and (cwithin c2) (mwithin foo))
; 		      (write 'c2-foo))
; 	      (before (and (cwithin c1) (mwithin bar))
; 		      (write 'c1-bar))
; 	      (before (and (cwithin c2) (mwithin bar))
; 		      (write 'c2-bar))
; 	      (before (and (cwithin c1) (mwithin baz))
; 		      (write 'c1-baz))
; 	      (before (and (cwithin c2) (mwithin baz))
; 		      (write 'c2-baz))))
; 	  7
; 	  '(c1-foo c2-foo c1-foo c1-bar c2-bar c1-baz c1-baz))


; (add-test 'mwithin02
; 	  "mwithin"
; 	  `((class main object
; 	      (method void init () (super init))
; 	      (method int main ()
; 		(send (new pos) compute)))
	    
; 	    (class pos object
; 	      (field int x)
; 	      (field int y)
; 	      (method void init () (fset! this x 1) (fset! this y 2) (super init))
; 	      (method int get-x () (fget this x))
; 	      (method int get-y () (fget this x))
; 	      (method int compute () (+ (send this get-x) (fget this y))))

; 	    (class enforcement object
; 	      (before (and (or (get x)                  ;want to write (fget *)
; 			       (get y))
; 			   (not (or (mwithin get-x)     ;want to write (mwithin get-*)
; 				    (mwithin get-y))))
; 		(write 'bad-get))))
; 	  3
; 	  '(bad-get))


#|


(add-test 'or01
	  "or pcd"
	  '((procedure void main ()
		       (write (fact 6)) (newline)
		       (write (fib  6)) (newline)
		       1)
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (procedure int fib ((int n))
		       (if (= n 0)
			   1
			   (if (= n 1)
			       1
			       (+ (fib (- n 1)) (fib (- n 2))))))
	    (before (and (or (call fact)
			     (call fib))
			 (args (int x)))
		    (write 'before) (write x) (newline)))
	  1
	  (append (append (merge (copies 7 'before)
				 (reverse (iota 7))
				 (copies 7 'NL))
			  '(720 NL))
		  (append (merge (copies 25 'before)
				 '(6 5 4 3 2 1 0 1 2 1 0 3 2 1 0 1 4 3 2 1 0 1 2 1 0)
				 (copies 25 'NL))
			  '(13 NL))))

(add-test 'not01
	  "not"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (and (call fact)
			 (not (cflowbelow (call fact))) ;cflow gives no jps
			 (args (int x)))
		    (write 'before) (write x) (newline)))
	  720
	  '(before 6 NL))

(add-test 'pwithin01a
	  "pwithin"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (pwithin void main ())
		    (write 'before) (newline)))
	  720
	  '(before NL before NL))

(add-test 'pwithin01b
	  "pwithin"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (and (pwithin void main ())
			 (call fact))
		    (write 'before) (newline)))
	  720
	  '(before NL))

(add-test 'pwithin02a
	  "pwithin"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (pwithin int fact (int))
		    (write 'before) (newline)))
	  720
	  (merge (copies 13 'before)
		 (copies 13 'NL)))

(add-test 'pwithin02b
	  "pwithin"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (and (pwithin int fact (int))       ;only picks up the pexecution jps
			 (not (call fact)))
		    (write 'before) (newline)))
	  720
	  (merge (copies 7 'before)
		 (copies 7 'NL)))

(add-test 'pwithin02c
	  "pwithin"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (and (pwithin int fact (int))      ;only picks up pcall jps
			 (call fact))
		    (write 'before) (newline)))
	  720
	  (merge (copies 6 'before)
		 (copies 6 'NL)))

(add-test 'pwithin03
	  "and pcd and pwithin"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (and (pwithin void main ())
			 (call fact)
			 (args (int m)))
		    (write 'pwithin) (write 'main) (write m) (newline)))
	  720
	  '(pwithin main 6 NL))

(add-test 'pwithin04
	  "and pcd and or pcd"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int n))
		       (if (= n 0) 1
			   (* n (fact (- n 1)))))
	    (before (and (or (and (pwithin void main ())
				  (call fact))
			     (and (pwithin int fact (int))
				  (call fact)))
			 (args (int x)))
		    (write x) (newline)))
	  720
	  (merge (reverse (iota 7))
		 (copies 7 'NL)))



(add-test 'proceed01
	  "capturing return values from proceed"
	  '((procedure void main () (fact 6))
	    (procedure int fact ((int x))
		       (if (= x 0) 1
			   (* x (fact (- x 1)))))
	    (around (and (call fact)
			 (args (int x)))
		    (let (((int result) 0))
		      (write x) (newline)
		      (set! result (+ 1 (proceed x)))
		      (write result) (newline)
		      result)))
	  2677
	  (append (merge (reverse (iota 7))
			 (copies 7 'NL))
		  (merge '(2 3 7 22 89 446 2677)
			 (copies 7 'NL))))


(add-test 'call01
	  "before and after method calls"
	  '((procedure void main () (let (((c1 o1) (new c1)))
				      (send o1 bar)))
	    
	    (class c1 object
		   (method int bar   () 1)
		   (method void init () 1))

	    (before (call bar)
		    (write 'before))
	    (after (call bar)
		   (write 'after)))
	  1
	  '(before after))

(add-test 'call02
	  "around method calls and target typematching"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2)))
				      (send o1 bar)
				      (send o2 bar)))
	    
	    (class c1 object
		   (method int bar   () 1)
		   (method void init () 1))

	    (class c2 object
		   (method int bar   () 2)
		   (method void init () 2))

	    (around (and (call bar)
			 (target (c1 c)))
		    (let (((int r) 0))
		      (write 'around-before)
		      (set! r (proceed c))
		      (write 'around-after)
		      r)))
	  2
	  '(around-before around-after))

(add-test 'call03
	  "around method calls and target inheritance"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2)))
				      (+ (send o1 bar)
					 (send o2 bar))))
	    
	    (class c1 object
		   (method int bar   () 1)
		   (method void init () 1))

	    (class c2 c1
		   (method int bar   () 2)
		   (method void init () 2))

	    (around (and (call bar)
			 (target (c1 c)))
		    (let (((int r) 0))
		      (write 'around-before)
		      (set! r (proceed c))
		      (write 'around-after)
		      r)))
	  3
	  '(around-before around-after around-before around-after))

(add-test 'call04
	  "method calls and args typematching"
	  '((procedure void main () (let (((c1 o1) (new c1)))
				      (send o1 bar 1)))
	    
	    (class c1 object
		   (method int bar   ((int x)) (+ x 1))
		   (method void init ()        1))

; 	    (before (and (call bar) ; args-pcd fails to find superclass of int
; 			 (args (foo f))
; 		    (write 'before-foo)
; 		    (write f))

	    (before (and (call bar)
			 (args (int i)))
		    (write 'before-int)
		    (write i)))
	  2
	  '(before-int 1))

(add-test 'call05
	  "method calls and args typematching"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((a1 a1) (new a1))
					  ((a2 a2) (new a2))
					  ((a1 a3) (new a3))) 
				      (+ (send o1 bar a1)
					 (send o1 bar a2)
					 (send o1 bar a3))))
	    
	    (class c1 object
		   (method int bar   ((a1 a)) (send a foo))
		   (method void init ()       1))

	    (class a1 object
		   (method int  foo  () 1)
		   (method void init () 1))
	    (class a2 object
		   (method int  foo  () 2)
		   (method void init () 2))
	    (class a3 a1
		   (method int  foo  () 3)
		   (method void init () 3))

	    (before (and (call bar)
			 (args (a1 a)))
		    (write 'before-a1))
	    (before (and (call bar)
			 (args (a2 a)))
		    (write 'before-a2))
	    (before (and (call bar)
			 (args (a3 a)))
		    (write 'before-a3)))
	  6
	  '(before-a1 before-a2 before-a3 before-a1))


(add-test 'mexecs01
	  "before and after method execution"
	  '((procedure void main () (let (((c1 o1) (new c1)))
				      (send o1 bar)))
	    
	    (class c1 object
		   (method int bar   () 1)
		   (method void init () 1))

	    (before (mexecution int c1 bar ())
		    (write 'before))
	    (after (mexecution int c1 bar())
		   (write 'after)))
	  1
	  '(before after))

(add-test 'mexecs02
	  "around method execution and target typematching"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2)))
				      (send o1 bar)
				      (send o2 bar)))
	    
	    (class c1 object
		   (method int bar   () 1)
		   (method void init () 1))

	    (class c2 object
		   (method int bar   () 2)
		   (method void init () 2))

	    (around (and (mexecution int c1 bar())
			 (target (c1 c)))
		    (let (((int r) 0))
		      (write 'around-before)
		      (set! r (proceed c))
		      (write 'around-after)
		      r)))
	  2
	  '(around-before around-after))

(add-test 'mexecs03
	  "around method execution and target inheritance"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2)))
				      (+ (send o1 bar)
					 (send o2 bar))))
	    
	    (class c1 object
		   (method int bar   () 1)
		   (method void init () 1))

	    (class c2 c1
		   (method int bar   () 2)
		   (method void init () 2))

	    (around (and (mexecution int c1 bar())
			 (target (c1 c)))
		    (let (((int r) 0))
		      (write 'around-before)
		      (set! r (proceed c))
		      (write 'around-after)
		      r)))
	  3
	  '(around-before around-after around-before around-after))

(add-test 'mexecs04
	  "method execution and args typematching"
	  '((procedure void main () (let (((c1 o1) (new c1)))
				      (send o1 bar 1)))
	    
	    (class c1 object
		   (method int bar   ((int x)) (+ x 1))
		   (method void init ()        1))

; 	    (before (and (mexecution int c1 bar (int)) ; args-pcd fails to find superclass of int
; 			 (args (foo f))
; 		    (write 'before-foo)
; 		    (write f))

	    (before (and (mexecution int c1 bar (int))
			 (args (int i)))
		    (write 'before-int)
		    (write i)))
	  2
	  '(before-int 1))

(add-test 'mexecs05
	  "method execution and args typematching"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((a1 a1) (new a1))
					  ((a2 a2) (new a2))
					  ((a1 a3) (new a3))) 
				      (+ (send o1 bar a1)
					 (send o1 bar a2)
					 (send o1 bar a3))))
	    
	    (class c1 object
		   (method int bar   ((a1 a)) (send a foo))
		   (method void init ()       1))

	    (class a1 object
		   (method int  foo  () 1)
		   (method void init () 1))
	    (class a2 object
		   (method int  foo  () 2)
		   (method void init () 2))
	    (class a3 a1
		   (method int  foo  () 3)
		   (method void init () 3))

	    (before (and (mexecution int c1 bar (a1))
			 (args (a1 a)))
		    (write 'before-a1))
	    (before (and (mexecution int c1 bar (a2))
			 (args (a2 a)))
		    (write 'before-a2))
	    (before (and (mexecution int c1 bar (a3))
			 (args (a3 a)))
		    (write 'before-a3)))
	  6
	  '(before-a1 before-a2 before-a3 before-a1))

(add-test 'target01
	  "target"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2))
					  ((c1 o3) (new c3)))
				      (+ (send o1 foo)
					 (send o2 foo)
					 (send o3 foo))))

	    (class c1 object
		   (method int foo   () 1)
		   (method void init () 1))

	    (class c2 object
		   (method int foo   () 2)
		   (method void init () 2))

	    (class c3 c1
		   (method int foo   () 3)
		   (method void init () 3))

	    (before (target (c1 c))
		    (write 'target-c1))
	    (before (target (c2 c))
		    (write 'target-c2))
	    (before (target (c3 c))
		    (write 'target-c3)))
	  6
	  '(target-c1 target-c2 target-c3 target-c1   ;mexecs of init
	    target-c1 target-c1 target-c2 target-c2   ;mcall mexec of foo
	    target-c3 target-c1 target-c3 target-c1)) ;mcall mexec of foo


(add-test 'this01
	  "this"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2))
					  ((c1 o3) (new c3)))
				      (+ (send o1 foo)
					 (send o2 foo)
					 (send o3 foo))))

	    (class c1 object
		   (method int foo   () 1)
		   (method void init () 1))

	    (class c2 object
		   (method int foo   () 2)
		   (method void init () 2))

	    (class c3 c1
		   (method int foo   () 3)
		   (method void init () 3))

	    (before (this (c1 c))
		    (write 'this-c1))
	    (before (this (c2 c))
		    (write 'this-c2))
	    (before (this (c3 c))
		    (write 'this-c3)))
	  6
	  '(this-c1 this-c2 this-c3 this-c1   ;mexecs of init
	    this-c1 this-c2 this-c3 this-c1)) ;mexecs of foo
	  
	  
 ;mexec join-points

'(add-test 'mwithin02
	  "mwithin in cflow"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c1 o2) (new c2)))
				      (+ (send o1 foo)
					 (send o2 foo))))
	    
	    (class c1 object
		   (method int foo   () 1)
		   (method void init () 1))

	    (class c2 c1
		   (method int foo   () (super foo))
		   (method void init () 2))

	    ;;
	    ;; except for testing, this is a crazy way to use cflow
	    ;;
	    (before (and (cflow (mwithin int c2 foo ())))
		    (write 'before1)))
	  2
	  '(before1 before1))

(add-test 'cwithin01
	  "this"
	  '((procedure void main () (let (((c1 o1) (new c1))
					  ((c2 o2) (new c2))
					  ((c1 o3) (new c3)))
				      (+ (send o1 foo)
					 (send o2 foo)
					 (send o3 foo))))

	    (class c1 object
		   (method int foo   () 1)
		   (method void init () 1))

	    (class c2 object
		   (method int foo   () 2)
		   (method void init () 2))

	    (class c3 c1
		   (method int foo   () (super foo) 3)
		   (method void init () 3))

	    (before (cwithin c1)
		    (write 'within-c1))
	    (before (cwithin c2)
		    (write 'within-c2))
	    (before (cwithin c3)
		    (write 'within-c3)))
	  6
	  '(within-c1 within-c2 within-c3 within-c1 within-c2 within-c3 within-c1))

(add-test 'news01
	  "new object join-points"
	  '((procedure void main () (let (((c1 o1) (new c1 1))
					  ((c2 o2) (new c2 2)))
				      (+ (send o1 foo o1)
					 (send o2 foo o2))))
	    
	    (class c1 object
		   (field int y)
		   (method int foo   ((c1 c))   (send c bar))
		   (method int bar   ()         (fget this y))
		   (method void init ((int ty)) (fset! this y ty)))

	    (class c2 c1
		   (field int x)
		   (method int bar   ()         (+ (fget this x) (super bar)))
		   (method void init ((int tx)) (super init (* 2 tx)) (fset! this x tx)))

 	    (before (news c1 (int))
 		    (write 'before))

	    (after (or (news c1 (int))
		       (news c2 (int)))
		   (write 'after)))
	  7
	  '(before after after))

(add-test 'news02
	  "change new args"
	  '((procedure void main () (let (((c1 o1) (new c1 0))
					  ((c2 o2) (new c2 2)))
				      (+ (send o1 foo o1)
					 (send o2 foo o2))))
	    
	    (class c1 object
		   (field int y)
		   (method int foo   ((c1 c))   (send c bar))
		   (method int bar   ()         (fget this y))
		   (method void init ((int ty)) (fset! this y ty)))

	    (class c2 c1
		   (field int x)
		   (method int bar   ()         (+ (fget this x) (super bar)))
		   (method void init ((int tx)) (super init (* 2 tx)) (fset! this x tx)))

 	    (around (and (news c1 (int))
			 (args (int x)))
		    (if (= x 0)
			(proceed 1)
			(proceed x))))
	  7
	  '())

(add-test 'news03
	  "replace new entirely"
	  '((procedure void main () (let (((c1 o1) (new c1 0))
					  ((c2 o2) (new c2 2))
					  ((c1 o3) (new c1 1)))
				      (+ (send o1 foo o1)
					 (send o2 foo o2)
					 (send o3 foo o3))))
	    
	    (class c1 object
		   (field int y)
		   (method int foo   ((c1 c))   (send c bar))
		   (method int bar   ()         (fget this y))
		   (method void init ((int ty)) (fset! this y ty)))

	    (class c2 c1
		   (field int x)
		   (method int bar   ()         (+ (fget this x) (super bar)))
		   (method void init ((int tx)) (super init (* 2 tx)) (fset! this x tx)))

 	    (around (and (news c1 (int))
			 (args (int x)))
		    (if (= x 0)
			(new c2 1)
			(proceed x))))
	  10
	  '())
|#

; (add-test 'mitch01
; 	  "weird times and fact"
; 	  '((class main object
; 	      (method void init () (super init))

; 	      (method int main () 
; 		(list (send this itimes 3 6)
; 		      (send this rtimes 3 6)
; 		      (send this fact 6)))

; 	      (method int engine ((int a) (int b)) (send this engine a b))


; 	      ;;
; 	      ;; Think of these two definitions as actually being defined in the
; 	      ;; aspects below.  That's what the '[xx...]', ala aspectj-mode for
; 	      ;; emacs, is showing you.
; 	      ;;
; 	      (method int itimes ((int xo) (int yo)) (send this engine xo 0));[iter...]
; 	      (method int rtimes ((int xo) (int yo)) (send this engine xo 0));[recu...]
; 	      (method int fact   ((int xo))          (send this engine xo 1));[fact...]
; 	      )



; 	    (class iterative-multiplication object ;really an aspect
; 	     ;(method (main itimes) ((int xo) (int yo)) (send this engine xo 0))

; 	      (around (and (cflow (and (call itimes)
; 				       (args (int xo) (int yo))))
; 			   (call engine)
; 			   (args (int x) (int acc)))
; 	       ;(newline) (write (list xo yo x acc))
; 		(if (zero? x)
; 		    acc
; 		    (proceed xo yo (- x 1) (+ acc yo))))
; 	      )

; 	    (class recursive-multiplication object ;really an aspect
; 	     ;(method (main rtimes) ((int xo) (int yo)) (send this engine xo 0))

; 	      (around (and (cflow (and (call rtimes)
; 				       (args (int xo) (int yo))))
; 			   (call engine)
; 			   (args (int x) (int ignore)))
; 	       ;(newline) (write (list xo yo x))
; 		(if (zero? x)
; 		    0
; 		    (+ (proceed xo yo (- x 1) ignore) yo)))
; 	      )

; 	    (class factorial object	    ;really an aspect
; 	     ;(method (main fact) ((int xo)) (send this engine xo 1))

; 	      (around (and (cflow (call fact))
; 			   (call engine)
; 			   (args (int x) (int acc)))
; 	       ;(newline) (write (list x acc))
; 		(if (zero? x)
; 		    acc
; 		    (proceed (- x 1) (* x acc))))
; 	      ))
; 	  '(18 18 720)
; 	  '())
