#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <2002-02-18 13:13:09 gregor>"))
  (display (string-append "tests=qb.scm:  " (substring time-stamp 13 29)))
  (newline))

;;; a smaller version of the figure program.  this can be tested with our
;;; QB implementation which uses O(e^N) algorithm for N language elements.

(define figure-program
  `((class Point object
      (field int x)
      (field int y)
      (method void init ((int x) (int y)) 
	(super init)
	(send this moveTo x y))
      (method void moveTo ((int x) (int y))
	(fset! this x x)
	(fset! this y y))
      (method int getx ()
	(fget this x))
      (method int gety ()
	(fget this y)))

    (class Line object
      (field Point p1)
      (field Point p2)
      (method void init ()
	(super init)
	(send this moveTo (new Point 0 0)  (new Point 0 0)))
      (method void moveTo ((Point new-p1) (Point new-p2))
        (fset! this p1 new-p1)
        (fset! this p2 new-p2))
      (method int length ()
	(let (((int dx) (- (send p2 getx) (send p1 getx))
	       (int dy) (- (send p2 gety) (send p2 gety))))
	  (+ (* dx dx) (* dy dy)))))))

(define query1
  (lambda (request)
    (cond ((eq? request 'to-string)
	   (lambda () (string-append
		       "?- class(C), name(C,Cname), "
		       "method(M), name(M,Mname), contains(C,M).")))
	  ((eq? request 'query-vars)
	   (lambda () '(c m cname mname)))
	  ((eq? request 'test?)
	   (lambda (env)
	     (let ((c (lookup-var 'c env))
		   (m (lookup-var 'm env))
		   (cname (lookup-var 'cname env))
		   (mname (lookup-var 'mname env)))
	       (and (class? c) (method? m) (contains? c m)
		    (eq? (class-cname c) cname) (eq? (method-mname m) mname))))))))

(define query2
  (lambda (request)
    (cond ((eq? request 'to-string)
	   (lambda () (string-append
		       "?- class(C), name(C,Cname), method(M), "
		       "name(M,Mname), contains(C,M), "
		       "member(M,[moveTo,length]).")))
	  ((eq? request 'query-vars)
	   (lambda () '(c m cname mname)))
	  ((eq? request 'test?)
	   (lambda (env)

	     (let ((c     (lookup-var 'c     env))
		   (m     (lookup-var 'm     env))
		   (cname (lookup-var 'cname env))
		   (mname (lookup-var 'mname env)))
	       (and (class? c)
		    (method? m)
		    (eq? (class-cname c) cname) (eq? (method-mname m) mname)
		    (memq mname '(moveTo length))
		    (contains? c m))))))))

(define test
  (lambda (query vars)
    (qb:weave (parse-program figure-program)
	      query vars)))

(define test1 (lambda () (test query1 '(cname mname))))
(define test2 (lambda () (test query1 '(mname cname))))
(define test3 (lambda () (test query2 '(cname mname))))
(define test4 (lambda () (test query2 '(mname cname))))
