#|<--                               90 columns                                        ->|#

(let ((time-stamp "Time-stamp: <2002-06-11 12:59:41 gregor>"))
  (display (string-append "weaver=compositor-ct.scm: " (substring time-stamp 13 29)))
  (newline))


;;; (revised version)
;;;
;;; X: composed program
;;; JPs: class, method and field declarations in X
;;;
;;; A: slices (or source declarations) and relationships
;;; IDa: matching-rules (equality over method signatures)
;;; EFFa: merging-rules (merge, override, etc.)
;;;
;;; (assumes only "merge-by-name" relationships)

(define compositor:weave;;HyperJ
  (lambda (slices relationships);;-> pgm

    ;; (Impl. note: you can't do this with fold.)

    (let loop ((pgm   (make-program '()))     
               (seeds (compute-seeds (flatmap slice-decls slices)))); candidates JPs
      (if (not (null? seeds))
          (let ((signature (all-match (car seeds) relationships)))  ;ID-A,B
            (if signature
                (let* ((jp (car seeds))                             ;this is now
                                                                    ;definitely a JP

                                                                    ;\ EFF-A,B is to 
                       (decl      (merge-decls jp relationships)))  ;  merge decls and
                  (loop (put-decl-to-pgm decl pgm signature)        ;/ put into pgm

                        (remove-subsets jp (cdr seeds))))           ;seeds that are 
                                                                    ;subsets of JP are
                                                                    ;now also done
                (loop pgm (cdr seeds))))
          pgm))))


(define all-match
  (lambda (decls relationships)
    (if (= (length decls) 1)
	(decl->signature (car decls))
	(let* ((pairs   (all-pairs decls))
	       (matches (map (lambda (pair)
			       (signature-match? (car pair) (cdr pair) relationships))
			     pairs)))
	  (and (every? (lambda (sig) (match-byname? (car matches) sig))
		       (cdr matches))
	       (car matches))))))

(define all-pairs
  (lambda (lst)
    (let loop ((lst lst))
      (if (null? lst)
	  '()
	  (append (map (lambda (2nd-elm) (cons (car lst) 2nd-elm))
		       (cdr lst))
		  (loop (cdr lst)))))))

'(define pairwise-every?
  (lambda (pred lst)
    (let every-for-first? ((lst lst))
      (or (null? lst)
	  (and (every? (lambda (elm) (pred (car lst) elm)) (cdr lst))
	       (every-for-first? (cdr lst)))))))

(define remove-subsets
  (lambda (set set-list)
    (remove*-if (lambda (other-set) (subset? other-set set))
	       set-list)))

(define compute-signature
  (lambda (decls ignore-relationships)
    (decl->signature (car decls))))

;; Here are some implicit orders in the returned list of seeds.
;; 1. set of classes should come earlier than sets of other decls
;;    (due to the simplified implementation of merge algorithm)
;; 2. the list  must be sorted in set-inclusion order; i.e.,
;;    when s_i and s_j are i'th and j'th element of the returned
;;    list, and s_j is a subset of s_i, then i < j.

'(define compute-seeds			    ; returns a list of seeds (subsets of
   (lambda (decls)			    ; declarations).  the list is ordered so 
     (sort-by superset>subset?		    ; that (1) supersets come earlier than their
	      (sort-by class>method&field?  ; subsets, and (2) class declarations come
		       (powerset decls))))) ; earlier than method and field declarations.

;; actual implementation
(define compute-seeds
  (lambda (decls)
    (append (remove-if null? (powerset (collect-if class-decl?  decls)))
	    (remove-if null? (powerset (collect-if method-decl? decls)))
	    (remove-if null? (powerset (collect-if field-decl?  decls))))))

;;;
;;; identifying matching source declarations to a jp
;;;

'(define-struct matching-decl (decl merge-rule))

'(define match-decl
  (lambda (relationships signature decl)
    (cond ((byname-relation? relationships)
	   (and (signature-match? (decl->signature decl) signature)
		(make-matching-decl decl
				    (relationships-merge relationships))))
	  ;; other matching rules should go below
	  ;; ...
	  (error 'match-decl "no matching mechanism for ~s" relationships))))

(define signature-match?
  (lambda (sig1 sig2 relationships)
    (let try-rest-rules ((rules (relationships-match-rules relationships)))
      (and (not (null? rules))
	   (or ((car rules) sig1 sig2)
	       (try-rest-rules (cdr rules)))))))

;;;
;;; merging declarations into a program
;;;

(define merge-decls
  (lambda (decls relationships)
    (let ((merger     (relationships-merge relationships))
	  (last-decl  (car (last decls))))
      (cond ((and (eq? merger 'merge) (class-decl? last-decl))
	     (make-class-decl (class-decl-cname last-decl)
			      (class-decl-sname last-decl)
			      '()))

	    ((and (eq? merger 'override) (class-decl? last-decl))
	     (make-class-decl (class-decl-cname last-decl)
			      (class-decl-sname last-decl)
			      '()))

	    ((and (eq? merger 'merge) (method-decl? last-decl))
	     (make-method-decl (member-decl-cname last-decl)
			       (method-decl-rtype last-decl)
			       (method-decl-mname last-decl)
			       (method-decl-params last-decl)
			       (make-body
				(flatmap (lambda (decl)
					   (body-exps (method-decl-body decl)))
					 decls))))

	    ((and (eq? merger 'override) (method-decl? last-decl))
	     (make-method-decl (member-decl-cname last-decl)
			       (method-decl-rtype last-decl)
			       (method-decl-mname last-decl)
			       (method-decl-params last-decl)
			       (make-body (body-exps (method-decl-body
						      last-decl)))))

	    ((and (eq? merger 'merge)    (field-decl? last-decl))
	     last-decl)

	    ((and (eq? merger 'override) (field-decl? last-decl))
	     last-decl)

	    (else (error 'merge-decls "unimplemented merge method ~a" merger))))))

(define put-decl-to-pgm
  (lambda (decl pgm sig)
    (cond ((class-decl? decl)
	   (make-program (cons decl (program-decls pgm))))
	  ((or (method-decl? decl) (field-decl? decl))
	   (replace-class sig pgm
			  (lambda (class-decl)
			    (add-member sig class-decl decl)))))))


;;;
;;; relationships:
;;;	e.g., mergeByName;
;;;	        override method concern1.bar with method concern2.baz;
;;;		equate method aaa with method bbb;
;;; (supports only mergebyname for now)

; ((merge (byname))
;  (merge (rename method (c1 m1) (c2 m2))))

(define-struct relationships (match-rules merge))

(define match-byname?
  (lambda (sig1 sig2)
    (or (and (class-signature? sig1)
	     (class-signature? sig2)
	     (eq? (signature-cname sig1) (signature-cname sig2))
	     sig1)
	(and (method-signature? sig1)
	     (method-signature? sig2)
	     (eq? (signature-cname sig1) (signature-cname sig2))
	     (eq? (signature-mname sig1) (signature-mname sig2))
	     sig1)
	(and (field-signature? sig1)
	     (field-signature? sig2)
	     (eq? (signature-cname sig1) (signature-cname sig2))
	     (eq? (signature-mname sig1) (signature-mname sig2))
	     sig1))))

(define match-rename?
  (lambda (args)			; args ::= `(class  (,cn1)     (,cn2))
					;        | `(method (,cn ,mn1) (,cn ,mn2))
					;        | `(field  (,cn ,fn1) (,cn ,fn2))
    (let ((renamer
	   (lambda (sig)
	     (let ((kind (car args)))
	       (cond ((eq? kind 'class)
		      (let ((cn1 (caadr args)) (cn2 (caaddr args)))
			(if (and (class-signature? sig)
				 (eq? cn1 (signature-cname sig)))
			    (make-class-signature cn2)
			    sig)))
		     ((eq? kind 'method)
		      (let ((cn1 (caadr args))  (mn1 (cadadr args))
			    (cn2 (caaddr args)) (mn2 (car (cdaddr args))))			   
			(if (and (method-signature? sig)
				 (eq? cn1 (signature-cname sig))
				 (eq? mn1 (signature-mname sig)))
			    (make-method-signature cn2 mn2)
			    sig)))

		     ((eq? kind 'field)
		      (let ((cn1 (caadr args))  (mn1 (cadadr args))
			    (cn2 (caaddr args)) (mn2 (car (cdaddr args))))			   
			(if (and (field-signature? sig)
				 (eq? cn1 (signature-cname sig))
				 (eq? mn1 (signature-mname sig)))
			    (make-field-signature cn2 mn2)
			    sig))))))))
      (lambda (sig1 sig2)
	(or (match-byname? (renamer sig1) sig2)
	    (match-byname? sig1           (renamer sig2))
	    )))))

(define merge-relation?
  (lambda (relationships)
    (eq? (relationships-merge relationships) 'merge)))

(define merge-by-name
  (make-relationships (list match-byname?) 'merge))

;;;  signature
;;;

(define decl->signature identity)

(define decl->cname
  (lambda (decl)
    (cond ((class-decl?  decl) (class-decl-cname decl))
	  ((field-decl?  decl) (member-decl-cname decl))
	  ((method-decl? decl) (member-decl-cname decl)))))

(define decl->name
  (lambda (decl)
    (cond ((class-decl?  decl) (class-decl-cname decl))
	  ((field-decl?  decl) (field-decl-fname decl))
	  ((method-decl? decl) (method-decl-mname decl)))))

(define signature-cname decl->cname)
(define signature-sname class-decl-sname)
(define signature-mname decl->name)
(define signature-rtype
  (lambda (member-decl)
    (cond ((method-decl? member-decl) (method-decl-rtype member-decl))
	  ((field-decl?  member-decl) (field-decl-type member-decl)))))
(define signature-params method-decl-params)
(define class-signature? class-decl?)
(define method-signature? method-decl?)
(define field-signature? field-decl?)

(define make-method-signature
  (lambda (cname mname)
    (make-method-decl cname 'dummy-rtype mname '() 'dummy-body)))

;;;
;;; put a declaration into a program
;;;

(define replace-class
  (lambda (signature pgm cdecl->cdecl)
    (make-program
     (map (lambda (decl)
	    (if (and (class-decl? decl)
		     (eq? (signature-cname signature) (decl->cname decl)))
		(cdecl->cdecl decl)
		decl))
	  (program-decls pgm)))))

(define add-member
  (lambda (signature class-decl member-decl)
    (let ((mdecl
	   (cond ((method-decl? member-decl)
		  (make-method-decl (signature-cname signature)
			     (signature-rtype signature)
			     (signature-mname signature)
			     (signature-params signature)
			     (method-decl-body member-decl)))
		 ((field-decl? member-decl)
		  (make-field-decl (signature-cname signature)
			    (signature-rtype signature)
			    (signature-mname signature))))))
    (make-class-decl (class-decl-cname class-decl)
		     (class-decl-sname class-decl)
		     (cons mdecl (class-decl-decls class-decl))))))
	   

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; (old version)
;;; [HJ-remodularize] 
;;; X1: slices
;;;	ID:	singature
;;;     effect: provide body of declaration
;;;
;;; C:  relationships
;;;     ID: signature...
;;;     effect: merge, override...
;;;
;;; Xf: target program
;;;     JPs: declarations

; (define HJ-remodular:weave;;HyperJ
;   (lambda (slices relationships);;-> pgm
;     (let ((decls (flatmap slice-decls slices)))            ; A-->X
;       (fold
;        (lambda (decl pgm)
; 	 (let ((matching-relations (lookup-relations decl relationships))) ;ID-B
; 	   (fold (lambda (m-relation pgm)
; 		   (merge-to-pgm decl pgm m-relation)
; 		   ;;     \
; 		   ;;      \----------------------------- effect-B: merge decl into jp
; 		   )
; 		 matching-relations
; 		 pgm)))
;        decls
;        (make-program '())))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; [HJ-slice]
;;; A:  source program
;;;     ID:     signature of method decl
;;;     effect: run this code on call
;;;
;;; B:  concern map (description of slices)
;;;     ID:     concern map LHS
;;;     effect: collect in slice
;;;
;;; X1: slices                  (the next stages A)
;;;     JPs: declarations
;;;
;;; NOTE:
;;; - means of ID & effecting for A are not used in the weaving process.
;;; - X has declarations from A under the structure of B.
;;;

(define HJ-slice:weave
  (lambda (pgm concern-map);;-> slices

    (let* ((decls (program->decls pgm))) ; A-->X
      (fold
       (lambda (decl slices)		 ; decl is jp
	 ;;                                /------------------- ID-B: name of slice
	 (add-decl-to-slice decl slices ; /			for this jp
			  (lookup-slice-name decl concern-map))
	 ;;       \
	 ;;        \------------------------------------------- effect-B: put jp
	 ;;						        into slice
	 )
       decls
       (make-empty-slices)))))

;;; NOTE: (fold f (list v1 v2 .. vn) z) = (f vn (...(f v2 (f v1 z))))

(define lookup-slice-name
  (lambda (decl concern-map-decls)
    (let loop ((decls concern-map-decls))
      (if (pcd-match? (concern-map-decl-pcd (car decls)) decl)
	  (concern-map-decl-slice-name (car decls))
	  (loop (cdr decls))))))

(define add-decl-to-slice
  (lambda (decl slices name)
    (cond ((null? slices)
	   (list (make-slice name (list decl))))
	  ((equal? name (slice-slice-name (car slices)))
	   (cons (make-slice name (cons decl (slice-decls (car slices))))
		 (cdr slices)))
	  (else
	   (cons (car slices)
		 (add-decl-to-slice decl (cdr slices) name))))))



;;; slice

(define-struct slice  (slice-name decls))

(define make-empty-slices (lambda () '()))

(define find-slice
  (lambda (slice-name slices)
    (find-if (lambda (slice) (eq? slice-name (slice-slice-name slice)))
             slices)))

;;; concern map

(define-struct concern-map-decls (decls))
(define-struct concern-map-decl (pcd slice-name))

(define pcd-match?
  (lambda (pcd decl)
    (cond ((eqv? pcd 'default) #t)        
	  ((eqv? (car pcd) 'class)
	   (and (class-decl? decl)  (eq? (cadr pcd) (decl->name decl))))
	  ((eqv? (car pcd) 'field)
	   (and (field-decl? decl)  (eq? (cadr pcd) (decl->name decl))))
	  ((eqv? (car pcd) 'method)
	   (and (method-decl? decl) (eq? (cadr pcd) (decl->name decl)))))))

(define program->decls
  (lambda (pgm)
    (let ((decls '()))
      (walk-program pgm (lambda (x)
			  (if (declaration? x) (set! decls (cons x decls)))
			  x))
      decls)))

(define parse-concern-map  ;a SIMPLE parser
  (lambda (specs)
    (map (lambda (spec)
	   (make-concern-map-decl (car spec) (cadr spec)))
	 specs)))

