#|<--                                92 columns                                         ->|#


(let ((time-stamp "Time-stamp: <2002-06-11 12:59:38 gregor>"))
  (display (string-append "weaver=pa-rt.scm: " (substring time-stamp 13 29)))
  (newline))

;;; This is nano-AJD, an excerpt of AJD.
;;;	JP:	 method calls
;;;     PCDs:    call, target, args, or, and
;;;     advice:  after

;;;
;;;
;;; A runtime weaving implementation of the aspectj dynamic join-point-model style of
;;; aop.  
;;;
;;; It builds on the language defined in base.  
;;;
;;; The join points are: <n dynamic join point kinds>
;;;
;;; The means of identifying join points is: pointcuts
;;;
;;; The means of affecting behavior at the join points is: before, after and around
;;; advice.
;;;
;;;
;;; Numerous features of AspectJ are missing, including:
;;;
;;;    dispatch on static type in pointcuts
;;;    issingleton, perthis, pertarget, percflow
;;;

;;; A: source program (classes)
;;;    means ID:     signature of method decl
;;;    means effect: run this code on call
;;; 
;;; B: source program (aspects)
;;;    means ID:     pointcut
;;;    means effect: run this code before/after any methods that join here
;;; 
;;; X: computation
;;;    JPs: calls

; (define PA:weave
;   (lambda (pgm);; -> computation
;     (let ((adecls (collect-advice-decls pgm))
; 	  (pgm    (remove-advice-decls  pgm)))
;       (set! *advice-decls* adecls)
;       (eval-program pgm))))

(define PA:weave
  (lambda (pgm);; -> computation
    (set! *advice-decls* (collect-advice-decls pgm))
    (eval-program        (remove-advice-decls  pgm))))

(define call-method
  (lambda (mname obj args)
    (let* ((jp          (make-call-jp mname obj args)) ;JPs are method calls in X
	   (method      (lookup-method jp))            ;ID-A
	   (adv-matches (lookup-advice jp)))           ;ID-B
      (execute-advice adv-matches jp	               ;effect-B: run the advice code
	(lambda ()
	  (execute-method method jp))))))              ;effect-A: run the method body

(define execute-advice			; we know there are only after advices
  (lambda (adv-matches jp thunk)
    (let ((result (thunk)))
      (for-each (lambda (adv-match)
                  (execute-one-advice adv-match jp))
                adv-matches)
      result)))

(define execute-one-advice
  (lambda (adv-match jp)
    (eval-body (advice-decl-body (adv-match-adv-decl adv-match))
	       (make-advice-env adv-match jp))))

(define *advice-decls* '())

(define-struct adv-match (adv-decl ptc-match))

(define lookup-advice
  (lambda (jp)
    (remove* #f
      (map (lambda (adecl)
	     (let* ((pcd       (advice-decl-pcd adecl))	   ; pcd -> ptc
		    (ptc-match (pointcut-matches pcd jp))) ; pcd -> ptc
	       (if (not ptc-match)
		   #f
		   (make-adv-match adecl ptc-match))))
	   *advice-decls*))))


			    ; can remove here when they obstacle normal execution. 
(define make-advice-env
  (lambda (adv-match tjp)
    (let* ((ptc-match (adv-match-ptc-match adv-match))
	   (ids       (ptc-match-ids  ptc-match))
	   (vals      (ptc-match-vals ptc-match)))
      (new-env (cons 'tjp ids) (cons tjp vals)))))



;;; ----------------------------------------------------------------------------
;;; Join Points
;;;
;;;

(define-struct call-jp (mname target args))


;;; ----------------------------------------------------------------------------
;;; Join point designation (Point Cut Designators)
;;;
;;;

;; we know there are only call-jps

(define pointcut-matches
  (lambda (ptc jp)			; pcd -> ptc
    (cond ((call-pcd? ptc)		; call-pcd* -> call-pointcut*
	   (and (eq? (call-pcd-mname ptc)
		     (call-jp-mname jp))
		(make-ptc-match '() '()))) ; or make-env ???
	  
	  ((target-pcd? ptc)
	   (make-ptc-match (list (target-pcd-id ptc))
			   (list (call-jp-target jp))))
	  
	  ((args-pcd? ptc)
	   (make-ptc-match (args-pcd-ids ptc)
			   (call-jp-args jp)))
	  
	  ((and-pcd? ptc)
	   (let loop ((ptcs       (and-pcd-pcds ptc))
		      (prev-match (make-ptc-match '() '())))
	     (if (null? ptcs)
		 prev-match
		 (let ((this-match (pointcut-matches (car ptcs) jp)))
		   (and this-match
			(loop (cdr ptcs)
			      (ptc-match-append prev-match this-match)))))))

	  ((or-pcd? ptc)
	   (let loop ((ptcs (or-pcd-pcds ptc)))
	     (and (not (null? ptcs))
		  (or (pointcut-matches (car ptcs) jp)
		      (loop (cdr ptcs)))))))))

(define-struct ptc-match (ids vals))

(define ptc-match-append
  (lambda (m1 m2)
    (make-ptc-match (append (ptc-match-ids  m1) (ptc-match-ids  m2))
		    (append (ptc-match-vals m1) (ptc-match-vals m2)))))

(define target-pcd-id
  (lambda (target-pcd)
    (param-pname (target-pcd-param target-pcd))))

(define args-pcd-ids
  (lambda (args-pcd)
    (map param-pname (args-pcd-params args-pcd))))

;;; ----------------------------------------------------------------------------
;;; Parsing and loading (eopl calls loading elaboration)
;;;
;;;

(set! parse-member-decl
  (around-wrap parse-member-decl
    (lambda (original-parse-member-decl enclosing-cname sexp)
      (or (maybe-parse-advice   enclosing-cname sexp)
          (maybe-parse-pointcut enclosing-cname sexp)
          (original-parse-member-decl enclosing-cname sexp)))))

(define maybe-parse-advice
  (lambda (enclosing-cname sexp)
    (and (list? sexp)
         (memq (car sexp) '(before after around))
         (make-advice-decl enclosing-cname (car sexp) (parse-pcd (cadr sexp))
           (parse-body (cddr sexp))))))

(define maybe-parse-pointcut
  (lambda (enclosing-cname sexp)
    (and (list? sexp)
         (memq (car sexp) '(pointcut))
         (make-named-pcd-decl enclosing-cname (cadr sexp) (parse-params (caddr sexp))
	   (parse-pcd (cadddr sexp))))))

(define collect-advice-decls
  (lambda (pgm)
    (let ((advice-decls '()))
      (walk-program pgm
		    (lambda (decl)
		      (if (advice-decl? decl)
			  (set! advice-decls (cons decl advice-decls)))
		      decl))
      (reverse advice-decls))))

(define remove-advice-decls
  (lambda (pgm)
    (walk-program pgm
		  (lambda (decl)
		    (if (advice-decl? decl)
			#f		;means erasure
			decl)))))

;;; ----------------------------------------------------------------------
;;; patches to some base functions

;; let the weaver run before eval-program
(define run
  (lambda (text)
    (if (program? text)
	(PA:weave text)
	(PA:weave (parse-program text)))))

(define lookup-method			; (lookup-method jp) or
  (around-wrap lookup-method		; (lookup mname cname)
    (lambda (original-lookup-method jp-or-mname . maybe-cname)
      (let ((len (length maybe-cname)))
	(case len
	  ((0) (original-lookup-method (call-jp-mname jp-or-mname)
				     (object-cname (call-jp-target jp-or-mname))))
	  ((1) (original-lookup-method jp-or-mname (car maybe-cname)))
	  (else (error 'lookup-method "more than two args ~s" maybe-cname)))))))

(define execute-method			; (execute-method method jp) or
  (around-wrap execute-method		; (execute-method method obj args)
    (lambda (original-execut-method method jp-or-obj . maybe-args)
      (case (length maybe-args)
	((0) (original-execut-method method
				     (call-jp-target jp-or-obj)
				     (call-jp-args jp-or-obj)))
	((1) (original-execut-method method
				     jp-or-obj
				     (car maybe-args)))
	(else (error 'execute-method "more than three arguments ~s" maybe-args))))))


;;;
;;; this is the weaver
;;;
;;; note that in AspectJ, the advice rules are somewhat more complex.  
;;; Two important differences are that:
;;;
;;;    - advice is sorted by precedence
;;;
;;;    - each proceed causes a new advice lookup
;;;



;;;
;;; The next 11 definitions (3 variable definitions and 8 wrappers) are, in
;;; effect, an aspect that makes join points and calls the weaver.  (I would
;;; have liked to put them before the preceding weaver code, but I can't do
;;; that with this wrapper implementation.)
;;;
;;; Note that these three dynamic variables are ONLY used within these 11
;;; definitions.  They are like a "lexically extent dynamically scoped" variable
;;; of these wrappers.
;;;
;;; Whatever you do, don't get the idea of not passing tjp to push-jp, thinking
;;; it could just get it out of *jp*  Consider what would happen if we removed
;;; cflow like features from this language.  *jp* would go away, but we would still
;;; need to be able to call push-jp.
;;;
;
;(aspect weaver-calling
;


;;;
;;; make-jp-with-context-and-dispatch is supposed to make the join point and then call
;;; dispatch-advice.  In its full-blown version, it handles the context information of
;;; 
;;;   jp-within
;;;   jp-stack-previous
;;;   jp-time-previous
;;;
;;; But a simpler version of it could just look like:
;;;
;;;   (define make-jp-with-context-and-dispatch
;;;     (lambda (kind name this target args thunk)
;;;       (push-jp (make-jp kind name this target args thunk))))
;;;
;;;

;;;
;;; this just makes sure there is an outermost (lowermost whatever) join point
;;;


;) ;end of weaver-calling aspect
