#|<--                               90 columns                                        ->|#

(let ((time-stamp "Time-stamp: <2002-06-11 12:59:41 gregor>"))
  (display (string-append "weaver=qb-ct.scm: " (substring time-stamp 13 29)))
  (newline))

(load "weaver=compositor-ct.scm")


;;; "group classes by method names"
;;;
;;; class(C), method(M), contains(C, M), name(M, MN), (EQUAL(MN, 'm1') || EQUAL(MN, 'm2'))  :: MN, C
;;;
;;;
;;;   A    B
;;;    \   ^
;;;     v /
;;;      X
;;;
;;; remember that X is the COMBINED thing.  so in an analytic system, like QJB, the result of the
;;; analysis will come from X, rather than vice-versa
;;;
;;;
;;; also try to stop forgetting that since A and X are often fairly isomorphic, its easy to confuse them
;;;
;;;
;;; A:  source program (class and method decls (or maybe text))
;;;  ID:     nested signatures
;;;  effect: define metaobject(s) (text)
;;;
;;; B: tree
;;;  ID:     query
;;;               the query is the pointcut like thing
;;;               the env is the matching-ptc like thing, in this case
;;;               a structured group of JPs that together match the query
;;;                 the metaobjects in env, are the JPs
;;;                 the ids in envs are like vals in matching-ptc
;;;  effect: fold envs into tree    (<- effect back into B, not on X)
;;;
;;; X: "the program" (metaobjects)
;;;    JPs:     class and method metaobjects

(define qb:weave
  (lambda (pgm query tree-vars) ;->tree
    (let* ((metaobjects (gather-metaobjects pgm))
	  (all-envs    (possible-envs (query-vars query) metaobjects))
	  (tree	       (make-empty-tree)))

      (fold (lambda (env tree)
	      (if (match? env query)
		  (let ((vals (map (lambda (var) (lookup-var var env))
				   tree-vars)))
		    (add-to-tree vals tree))
		  tree))			  ; effect-B: put into tree
	    all-envs
	    tree))))
'
(define QBqb:weave
  (lambda (pgm query tree-vars);->tree
    (let* ((metaobjects (elaborate-program pgm))		; this is A-->X
           (all-envs    (possible-envs (query-vars query) metaobjects))
           (tree        (make-empty-tree)))		; this is B
      (for-each (lambda (env)
                  (if (match? env query)		; IdB
                    (let ((vals (map (lambda (var) (lookup-var var env))
                                     tree-vars)))
                      (add-to-tree! vals tree))))			; EffB
                all-envs)
      tree)))



;;; (add-to-tree (cons v1 vs) 
;;;              (make-tree v0 (list ... (make-tree v1 tN) ...)))
;;; => (make-tree v0 (list ... (add-to-tree vs (make-tree v1 tN)) ...))

(define add-to-tree
  (lambda (vals node)
    (if (null? vals)
	node
	(letrec ((loop
		  (lambda (subtrees)
		    (cond ((null? subtrees)
			   (cons (add-to-tree (cdr vals)
					      (make-tree (car vals) '()))
				 '()))

			  ((equal? (tree-val (car subtrees)) (car vals))
			   (cons (add-to-tree (cdr vals)
					      (car subtrees))
				 (cdr subtrees)))

			  (else
			   (cons (car subtrees)
				 (loop (cdr subtrees))))))))

	  (make-tree (tree-val node)
		     (loop (tree-subtrees node)))))))

;;; Tree ::= <Val, Tree*>

(define make-tree
  (lambda (val subtrees)
    (cons val subtrees)))
(define tree-val car)
(define tree-subtrees cdr)
(define make-empty-tree
  (lambda () (make-tree '*root* '())))

;;;

(define gather-metaobjects
  (lambda (pgm)
    (set! *classes* (init-classes))
    (elaborate-class-decls! (program-decls pgm))
    (append *classes* (apply append (map class-methods *classes*)))
;     (let ((classes (elaborate-class-decls (program-decls pgm))))
;       (append classes ))
    ))

;;;
;;; this is an abstraction of the logic engine
;;;
(define all-matching-envs
  (lambda (query vals)
    ;;
    ;; since we don't have a real query engine, we have to do some hacking
    ;; in here
    ;;
    (set! vals (append vals (all-metaobject-names vals)))
    (fold (lambda (env envs)
	    (let ((matching-envs (match? query env)))
	      (if matching-envs
		  (fold add-to-envs matching-envs envs)
		  envs)))
	  (possible-envs (query-vars query) vals)
	  empty-envs)))


(define add-to-envs cons)
(define empty-envs '())


(define possible-envs
  (lambda (vars vals)
    (set! vals (append vals (all-metaobject-names vals)))
    (let bind-a-var ((unbound-vars vars)
		     (bound-vars '())
		     (bound-vals  '()))
      (if (null? unbound-vars)
	  (list (new-env bound-vars bound-vals))
	  (flatmap (lambda (val)
		     (bind-a-var (cdr unbound-vars)
				 (cons (car unbound-vars) bound-vars)
				 (cons val                bound-vals)))
		   vals)))))

(define all-metaobject-names
  (lambda (metaobjects)
    (remove-duplicates
     (map (lambda (mo) (cond ((class? mo) (class-cname mo))
			     ((method? mo) (method-mname mo))))
	  metaobjects))))

(define query-vars			;a list of variables
  (lambda (query)			;in a query
    ((query 'query-vars))))

(define match?				;test the query under a partial 
  (lambda (env query)			;environment, and returns a list
    ((query 'test?) env)))		;of (complete) envs that satisfies.

;; the following two functions are used by queries as primitives 

(define contains?
  (lambda (container element)
    (and (class? container)
	 (or (and (method? element)
		  (eq? (method-cname element) (class-cname container) ))))))

(define lookup-var
  (lambda (var env)
    (deref (apply-env-ref env var))))
