#|<--                               90 columns                                        ->|#

(let ((time-stamp "Time-stamp: <2002-09-17 10:15:46 Hidehiko Masuhara>"))
  (display (string-append "weaver=travex-rt.scm: " (substring time-stamp 13 29)))
  (newline))

(load "weaver=trav-rt.scm")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; call visitors by executing methods with target's class name
;;;
;;; this implementation calls methods of visitor depending on the
;;; classes of targets.  this amounts to the mechanisms to call
;;; overloaded methods in DJ.  when a visitor has (method void visit-C
;;; ((C o)) ...), then the method is called when the traversal arrives
;;; objects subtypes of C.

;; (call-visitor v o) executes method visitor-C0, visitor-C1, ... of v
;; with o as an argument, where C0 is the dynamic class of o, and Cn+1
;; is the superclass of Cn.
(define call-visitor
  (lambda (visitor obj)
     (let ((mnames (visitor-mnames obj)))
       (for-each (lambda (mname)
 		  (if (has-method? (object-cname visitor)
 				   mname)
 		      (call-method mname
 				   visitor
 				   (list obj))))
 		mnames))))

(define visitor-mnames
  (lambda (obj)
    (case (object-cname obj)
      ((primitive) '())			; ignore primitives
      (else
       (map visitor-mname (object->ancestor-cnames obj))))))

(define visitor-mname
  (lambda (cname)
    (string->symbol
     (apply string-append
	    (map symbol->string (list 'visit- cname))))))

(define object->ancestor-cnames
  (lambda (object)
    (cname-ancestors (object-cname object))))

(define has-method?
  (lambda (cname mname)
    (find-if (lambda (method) (eqv? (method-mname method) mname))
	     (cname->methods cname))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; call visitors by executing advice 
;;;
;;; to define behavior at each object with advice, instead of visitor
;;; methods, we need to (1) add a new kind of join point and lookup
;;; and execute advice upon arrival (as seen in ajdex), and (2) define
;;; a new kind of pointcut to identify traversal join points (again,
;;; as seen in ajdex).
;;;
;;; this allows the programmer to control whether to continue traversal
;;; by writing around around, like this:
;;;     (advice (and (traversal counter)
;;;                  (target (figure-element o)))
;;;             (send counter addCounter))
;;;
;;; "crossing"-like features, which gives control the behavior based
;;; on the field names traversing, can be added (but not done) by:
;;; (1) extend traverse-jp structure to include fname,
;;; (2) extend arrive to receive fname, and to store it in jp, 
;;; (3) extend arrive-1 to call arrive with fname as well, and
;;; (4) extend pointcut-matches so that crossing pointcut matches to
;;;     traversal jps with the same fnames.

(load "weaver=paex-rt.scm")				  ;uses AJD's mechanism

(define-struct (traversal-jp struct:jp) (target visitor)) ;declare new kind of jp

(define trav:weave
  (lambda (trav-spec root visitor)
    (letrec ((arrive					         ;wrap original arrive to
	      (lambda (obj path)			         ;lookup/execute advice
		(let* ((jp (make-traversal-jp .stack-previous-jp.
					      obj
					      visitor))
		       (adv-matches (lookup-advice jp)))
		  (fluid-let ((.stack-previous-jp. jp))
		    (execute-advice adv-matches jp
		      (lambda () (arrive-1 obj path)))))))

	     (arrive-1					         ;this is originally what
	      (lambda (obj path)			         ;arrive does

		(call-visitor visitor obj)		         ;(for old examples; may
								 ;be deleted)
		(for-each (lambda (fname)
			    (let* ((next-obj   (get-field fname obj))
				   (next-cname (object-cname next-obj))
				   (next-path  (extend-path path next-cname)))
			      (if (match? next-path trav-spec)
				  (arrive next-obj next-path))))
			  (object->fnames obj)))))

      (arrive root (make-path (object-cname root))))))

(set! pointcut-matches
  (around-wrap pointcut-matches
    (lambda (other-pointcut-matches ptc jp)
      (cond ((traversal-pcd? ptc)			             ;traversal ptc 
	     (and (traversal-jp? jp)				     ;matches traversal
		  (make-ptc-match (list (traversal-pcd-name ptc))    ;jp, and binds 
				  (list (traversal-jp-visitor jp)))));visitor
	    ((target-pcd? ptc)
	     (and (target-jp? jp)
		  (subclass? (object-cname (jp-target jp))	     ;type check is
			     (target-pcd-type ptc))		     ;needed
		  (make-ptc-match (list (target-pcd-id ptc))
				  (list (jp-target jp)))))
	    (else
	     (other-pointcut-matches ptc jp))))))

(define target-pcd-type
  (lambda (ptc) (param-type (target-pcd-param ptc))))

(define target-jp?			;this may interfere existing adivce with
  (lambda (jp) (or (call-jp? jp)	;unquantified target ptcs
		   (get-jp? jp)
		   (traversal-jp? jp)))) 

(define jp-target
  (lambda (jp)
    (cond ((call-jp? jp)      (call-jp-target jp))
	  ((get-jp?  jp)      (get-jp-target  jp))
	  ((traversal-jp? jp) (traversal-jp-target jp))
	  (else		      (error 'jp-target "not appropriate jp ~s" jp)))))
